import os
import time
import typing
import uuid
from abc import ABC
from functools import wraps
from typing import Optional, Iterator, List, Dict, Union
import boto3
from boto3 import Session as Boto3Session
from cachetools import TTLCache, cached
import logging
from collections import defaultdict
from dateutil.parser import isoparse
from datetime import datetime
try:
    # these are only present for services, not for the SDK, and we do not wish to impose these in requirements.txt
    from flask import g
    from injector import inject
    from pymemcache import HashClient
except ModuleNotFoundError:
    # set up an Identity decorator for @inject
    def inject(function): # type: ignore
        def wrapper(*args, **kwargs):
            return function(*args, *kwargs)
        return wrapper


try:
    # these are only present for services, not for the SDK, and we do not wish to impose these in requirements.txt
    from ..providers.memcache_provider import memcached_method
except ModuleNotFoundError:
    def memcached_method(cache_getter, cache_func_name ,key=None, ttl=None, force_cache_refresh=False): # type: ignore
        def decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            return wrapper
        return decorator

from ..handlers.s3_partition_handler import get_partitions_in_filepath
from ..adapters.s3_io_adapters import S3File
from ..services.abstract_backend_handler_service import AbstractBackendDatasetHandler
from ..services.exceptions import UnableToAccessDataset

logger = logging.getLogger(__name__)

duration_seconds: int = int(os.environ.get(
    'ASSUME_ROLE_DURATION_SECONDS',
    '3600' # ideally we want this a lot longer, or specific to the bucket
))

S3PathDetails = dict


def _shorten(common_prefix: str, prefix: str):
    """
    Path is made by stripping the common prefix of the prefix
    and the delimiter, which S3 returns.

    :param common_prefix: e.g. abc/as_of_date=2019/
    :return: as_of_date=2019
    """

    top_level_path = common_prefix
    if common_prefix.startswith(prefix):
        top_level_path = common_prefix[len(prefix):]
    if common_prefix.endswith('/'):
        top_level_path = top_level_path[:-1]
    return top_level_path


def is_hidden_based_on_path_and_content(path: str, content_type: str) -> bool:
    """
    [DL-4545][DL-4536][DL-5209] Do not read into files or
    directories that are cruft from Spark which Spark will ignore
    on read, e.g. files/dirs starting with `.` or `_` are hidden
    to Spark.

    Skip as_of_*=latest as it is a Spark temporary folder.

    :param content_type: Structured/Unstructured
    :param path: e.g. as_of_date=2019, .latest, as_of_date=latest
    :return: True if the path matches the criteria of a hidden file.
    """
    return (
            path.startswith('.') or
            path.startswith('_') or
            (
                    path == 'metadata' and
                    content_type != "Unstructured"
            ) or
            (
                    path.startswith('as_of_') and
                    path.endswith('=latest')
            )
    )


class S3DatasetHandler(AbstractBackendDatasetHandler, ABC):

    """
    This is an abstract implementation, providing common methods for S3Datasets. The application itself should choose how
    to extend common methods specific for its uses, and any common ones across repos should live here
    """

    @inject
    def __init__(
        self,
        ttl_cache: TTLCache,
        hash_client: 'HashClient',
    ):
        """
        Constructor.

        :param tlc_cache: tbc
        :param hash_client: Used for caching responses from S3.
        """
        logger.debug('DatasetS3Handler constructor')
        super().__init__(ttl_cache=ttl_cache, hash_client=hash_client)

    def __del__(self):
        """Destructor."""
        logger.debug('Destructor called on DatasetS3Handler')

    def create_session(self, role_arn: Optional[str]) -> Boto3Session:
        """
        Assume the role using the role arn and the secret (in the environment
        variables for this environment).

        See https://stackoverflow.com/a/45834847

        The assuming the role takes about 0.3 seconds so this becomes
        expensive, especially for Big Data tools that make many recursive calls.

        :param role_arn: Role to assume.
        :return: A Boto3 Session which used the role_arn to assume the
        role, allowing access to the data on S3. It can automatically refresh
        when it reaches the expiry time.
        """

        @cached(cache=self.ttl_cache)
        def _create_session(_role_arn: Optional[str]) -> Boto3Session:

            # todo - dont think we need to check - as we just wont have access later
            # if not _role_arn:
            #     raise UnableToAuthenticateToS3(
            #         'AWS account has no RoleArn data. This means '
            #         'the business line which owns this AWS account, '
            #         'is still using an insecure way of authenticating to S3.'
            #     )

            # This is expensive as assume role talks to AWS, takes maybe
            # 0.3 seconds.

            # Refreshing session had memory leaks, so instead make a non-refreshing session.
            # https://stackoverflow.com/a/56262329

            # Duration can be up to to the maximum session duration setting for the role. This
            # setting can have a value from 1 hour to 12 hours. Let us try asking for the maximum
            # as we may have long running `dataframe` calls.
            # https://docs.aws.amazon.com/STS/latest/APIReference/API_AssumeRole.html
            # If you exceed the max for the role, you will get this exception message:
            # An error occurred (ValidationError) when calling the AssumeRole operation: The
            # requested DurationSeconds exceeds the MaxSessionDuration set for this role.

            # Re-use this request's request_id so if anyone cared to they could track the
            # request across services all the way to AWS.

            try:
                # this is for testing with localstack, changing the endpoint from real s3
                args = {}
                if os.environ.get("AWS_STS_URL", None):
                    args["endpoint_url"] = os.environ["AWS_STS_URL"]

                credentials = boto3.client("sts", **args).assume_role(
                    RoleSessionName=g.request_id if hasattr(g, "request_id") else str(uuid.uuid4()),
                    RoleArn=_role_arn,
                    DurationSeconds=duration_seconds
                )

                # Translate the keys from PascalCase to snake_case.
                return Boto3Session(
                    aws_access_key_id=credentials["Credentials"]["AccessKeyId"],
                    aws_secret_access_key=credentials["Credentials"]["SecretAccessKey"],
                    aws_session_token=credentials["Credentials"]["SessionToken"]
                )
            except Exception as e:
                logging.warning(e)
                raise e

        return _create_session(_role_arn=role_arn)

    def common_prefixes_on_s3(
        self,
        resource,
        bucket,
        prefix: str,
        is_top_level_call: bool,
    ) -> List[Dict]:
        """
        Checks to see if the common prefixes in cache match the common prefixes
        on S3 now. Returns the refreshed list of common prefixes.

        :param resource:
        :param bucket:
        :param prefix:
        :param is_top_level_call:
        :return: A list of the most recent common prefixes object summaries.
        """

        logger.debug(
            'Fetch common prefixes (likely the `as_of` partition) from '
            'cache',
            extra={
                'Bucket': bucket.name,
                'Prefix': prefix,
                'is_top_level_call': is_top_level_call,
            }
        )
        return list(
            self.get_common_prefixes(
                resource=resource,
                bucket=bucket,
                prefix=prefix,
                force_cache_refresh=is_top_level_call,
            )
        )

    def get_common_prefixes(
        self,
        resource,
        bucket,
        prefix,
        force_cache_refresh: bool,
    ) -> Iterator[Dict]:
        """
        Abstract over the S3 location and return a list of IO objects.

        :param prefix:
        :param bucket:
        :param resource:
        :param force_cache_refresh: When False, will use standard cache
            behaviour of checking if a value is in cache, then return from
            cache, else  make a real call. When True, will make a real call
            and replace the values in the cache.

        :return: Return type is really an iterator of s3.ObjectSummary, but
            s3.ObjectSummary is created by reflection so we cannot use type
            hinting.
        """
        logger.debug(
            'get_common_prefixes - start paginating s3 list from prefix to '
            'delimiter...',
            extra={
                'Bucket': bucket.name,
                'Prefix': prefix,
                'force_cache_refresh': force_cache_refresh,
            }
        )

        list_objects_func = memcached_method(
            lambda *_: self.hash_client,
            force_cache_refresh=force_cache_refresh,
            cache_func_name='list_objects_v2',
        )(resource.meta.client.list_objects_v2)

        delimiter = '/'
        start_after = None  # Used for pagination of S3 results.
        max_loops = 1000000  # Used to prevent infinite loops.
        i = 0

        while i < max_loops and True:
            i += 1  # Counter to prevent infinite loops.
            response: typing.Dict[str, str]

            if start_after:
                logger.debug(
                    'next page of s3 list results...',
                    extra={
                        'Bucket': bucket.name,
                        'Prefix': prefix,
                        'Delimiter': delimiter,
                        'start_after': start_after,
                        'force_cache_refresh': force_cache_refresh,
                    }
                )
                response = list_objects_func(
                    Bucket=bucket.name,
                    Prefix=prefix,
                    Delimiter=delimiter,
                    StartAfter=start_after,
                )
            else:
                # Initial loop
                logger.debug(
                    'start paginating s3 list...',
                    extra={
                        'Bucket': bucket.name,
                        'Prefix': prefix,
                        'Delimiter': delimiter,
                        'start_after': start_after,
                        'force_cache_refresh': force_cache_refresh,
                    }
                )
                response = list_objects_func(
                    Bucket=bucket.name,
                    Prefix=prefix,
                    Delimiter=delimiter,
                    # No StartAfter param in initial loop.,
                )

            response_as_dict: Dict = dict(response)

            # Remove the response metadata from the S3 list response as it
            # is different between requests because it contains the datetime
            # of the request and a unique request ID.
            if 'ResponseMetadata' in response_as_dict:
                del response_as_dict['ResponseMetadata']

            yield response_as_dict

            # Type could be horrible as it comes from a response.
            common_prefixes: Union[
                str,
                Dict[str, str],
                List[Dict[str, str]]
            ] = response.get('CommonPrefixes', [])

            last_prefix = None
            if common_prefixes:
                if isinstance(common_prefixes, str):
                    logger.warning(
                        'common prefix came back as a single string '
                        'instead of a list of dict which was unexpected '
                        'but was handled'
                    )
                    last_prefix = common_prefixes
                if isinstance(common_prefixes, Dict):
                    logger.warning(
                        'common prefix came back as a single dict '
                        'instead of a list of dict which was unexpected '
                        'but was handled',
                        extra={
                            'response first N chars': str(response)[:1000]
                        }
                    )
                    last_prefix = common_prefixes.get(
                        'Prefix',
                        None
                    )
                elif isinstance(common_prefixes, List):
                    # Example of common_prefixes structure:
                    #  [
                    #     {
                    #         "Prefix":
                    #             "short_code/as_of_date=2020-09-22/"
                    #     },
                    #     {
                    #         "Prefix":
                    #             "short_code/as_of_date=2020-09-23/"
                    #     }
                    # ]
                    for current_common_prefix in common_prefixes:
                        last_prefix = current_common_prefix.get(
                            'Prefix',
                            None
                        )
                else:
                    logger.error(
                        'type error for common_prefixes',
                        extra={
                            'common_prefixes': common_prefixes,
                            'type common_prefixes': type(common_prefixes),
                            'response first N chars': str(response)[:1000]
                        }
                    )
                    raise Exception(
                        'type error for common_prefixes - '
                        f'type: {type(common_prefixes)}, '
                        f'common_prefixes: {common_prefixes}'
                    )
            else:
                logger.debug(
                    'Zero common_prefixes',
                    extra={
                        'response first N chars': str(response)[:1000],
                    }
                )
                last_prefix = None

            if response.get('IsTruncated', False) and last_prefix:
                start_after = last_prefix
                logger.debug(
                    'Truncated response, so more results on the next '
                    'page',
                    extra={
                        'start_after': start_after,
                        'response first N chars': str(response)[:1000],
                    }
                )
            else:
                break

    # nonstatic
    def get_s3_list(
        self,
        resource,
        bucket_name: str,
        prefix: str,
        absolute_path: bool
    ) -> Iterator[S3PathDetails]:
        """
        S3 list everything in this bucket under this prefix.

        :param bucket:
        :param prefix: Tell s3-proxy/AWS S3 the exact folder from the S3 structure that you want to
          list inside:
          e.g. dataset-short-code/as_of_date=2019, or dataset-short-code/as_of_date
        :param absolute_path: True returns absolute path to the file on S3 proxy.
        :param bucket_name: The organisation short code is necessary as dataset short
          codes are NOT unique across different organisations.
        :return: S3 proxy path and size.
        """

        def __to_s3_dict(
                object_summary,
                absolute_path: bool,
                bucket_name: str,
        ) -> S3PathDetails:
            if absolute_path:
                key = f"s3://{bucket_name}/{object_summary.key}"
            else:
                key = object_summary.key

            return {
                'Key': key,
                'Size': object_summary.size,
                'LastModified': object_summary.last_modified,
                'ETag': object_summary.e_tag
            }


        logger.debug(f"Get S3 list for prefix '{prefix}'.")

        bucket = resource.Bucket(bucket_name)

        object_summaries = bucket.objects.filter(
            # Prefix searches for exact matches and folders
            Prefix=prefix
        )

        # Convert each object_summary into an S3 proxy path.
        for object_summary in object_summaries:
            if not object_summary.key.endswith('/') and int(object_summary.size) != 0:
                yield __to_s3_dict(
                    object_summary=object_summary,
                    absolute_path=absolute_path,
                    bucket_name=bucket_name
                )
            else:
                logger.debug(
                    'Skip zero size S3 path',
                    extra={
                        'key': object_summary.key,
                        'size': object_summary.size,
                    }
                )

    def get_s3_list_filter(
        self,
        resource,
        bucket_name: str,
        prefix: str,
        absolute_path: bool,
        prefix_filter_fn: typing.Callable[[Iterator[str]], Iterator[str]]
    ) -> typing.Generator[S3PathDetails, None, None]:
        """
        Top-level common prefix filtering. We need to prevent reading into
        hidden folders. In the case of Bilateral the latest folders are
        growing linearly and for us to do an iterative list and post filter
        out folders we do not want is taking too long on prod. We need to
        filter out at the common prefix level!

        :param resource:
        :param bucket:
        :param prefix: e.g. abc/as_of_date=2019, abc/as_of_date
        :param content_type: Structured/Unstructured
        :param load_type: Either None, "Incremental Load" or "Full Load". None
            should have the same behaviour as "Incremental Load".
        :param absolute_path: True returns absolute path to the file on S3 proxy.
        :param bucket_name:
        :return: S3 proxy path and size.
        """

        bucket = resource.Bucket(bucket_name)
        unique_common_prefixes: typing.Set[str] = set([])

        '''
         force_cache_refresh=True -- this is present simply because we do not have memcached on SDK - 
         we can remove and it will still do the real call, but lets be explicit
        '''
        common_prefixes_aws_responses: Iterator[Dict] = self.get_common_prefixes(
            resource=resource,
            bucket=bucket,
            prefix=prefix,
            force_cache_refresh=True
        )

        for aws_response in common_prefixes_aws_responses:
            common_prefixes = aws_response.get('CommonPrefixes', [])
            for cp in common_prefixes:
                unique_common_prefixes.add(cp['Prefix'])

        unique_non_hidden_common_prefixes = prefix_filter_fn(iter(unique_common_prefixes))

        for non_hidden_common_prefix in unique_non_hidden_common_prefixes:
            # To get a list of the contents call s3-proxy with
            # unique prefix plus partition
            # e.g. for
            #   abc/as_of_date
            # the S3 list could return:
            #   abc/as_of_date=2019/file.parquet
            #   abc/as_of_date=2020/file.parquet
            # why dont we cache this call?
            yield from self.get_s3_list(
                resource=resource,
                bucket_name=bucket_name,
                prefix=non_hidden_common_prefix,
                absolute_path=absolute_path
            )

    @staticmethod
    def convert_list_to_file_handles(
        resource,
        bucket_name,
        paths: Iterator[S3PathDetails]
    ) -> Iterator[S3File]:
        """
        List using caching. The intent is that this returns files for the
        `dataframe` and `partitions` endpoints, so we will be removing
        empty files and removing directories. This is NOT the behaviour you
        want for the S3 proxy, where you will want to view all of the files.

        :param resource:
        :param bucket:
        :param prefix:
        :param delimiter: Use with the `prefix` to do matching of path up to
            the delimiter.
        :param filter_out_directories: If True, remove any directories
            from the list so that it only returns files.
        :return: List of S3 files.
        """
        # There are two kinds of data in a datafile.files array. Absolute
        # paths. i.e. s3://bucket/prefix/userdata1.parquet and pointers
        # to folders  i.e s3://bucket/prefix/ which may contain part files
        # ending in 0001...000n. OR they may just contain general data.

        key: S3PathDetails
        try:
            for key in paths:
                # n.b. S3PathAndSize must be non-absolute for this method to work! (absolute_path=False)
                #
                yield S3File(
                    key_dict=key,
                    bucket_name=bucket_name,
                    get_object=resource.meta.client.get_object
                )
        except StopIteration:
            logger.debug("No more files in paths")

    @staticmethod
    def get_most_recent_common_prefix(
        common_prefixes: Iterator[str],
    ) -> List[str]:
        """
        This filtering logic only applies to Structured datasets. For a
        Structured dataset, we expect the first common prefix on S3 to be an
        `as_of_` representing when the data was published. It is expected to be
        in ISO 8601 format as this can be parsed into a Python DateTime.

        Return only the most recent `as_of` common prefix.

        :param common_prefixes: List of common prefixes for this dataset on S3.
        :return:  The most recent common_prefix for this dataset on S3. We return
            a collection because there is an edge case where there are common
            prefixes on S3 that do not have as_of, in which case
        """

        as_of_to_common_prefix: dict = defaultdict(list)
        as_of_warning_shown = False

        # Accumulate the `as_of_` dates from the common prefixes.
        for common_prefix in common_prefixes:
            partitions = {
                k.lower(): v for k, v in dict(
                    get_partitions_in_filepath(common_prefix)
                ).items()
            }

            # Partition cannot be filtered out if does not have an
            # `as_of_`. Get the first instance of a key that begins
            # with `as_of_`.
            as_of_key = next(
                (k for k, v in partitions.items()
                 if k.startswith('as_of_')),
                None
            )

            if not as_of_key:
                # There is not an `as_of_` in this common prefix!
                # Only show the message the first time, otherwise it will
                # spam the logs.
                if not as_of_warning_shown:
                    logger.warning(
                        'as_of_ not in partitions but load type is full, '
                        'so not filtering out this common prefix but will '
                        'consider it as having the oldest as_of_ possible.',
                        extra={
                            'common_prefix': common_prefix,
                        }
                    )
                    as_of_warning_shown = True

                as_of_to_common_prefix[
                    datetime.min
                ].append(common_prefix)
            else:
                # For a `load_type` = `Full Load` dataset, we only want
                # to return the data in the most recent `as_of_`
                # directory. We accumulate a dict of `as_of_` to
                # common_prefix, then at the end we know what the most recent
                # `as_of_` is and only return that common prefix.
                as_of_to_common_prefix[
                    isoparse(partitions[as_of_key])
                ].append(common_prefix)

        if as_of_to_common_prefix:
            # Return only the most recent common prefix.
            # To sort, we have to action the iterator and therefore the type
            # changes to a List.
            most_recent_as_of = sorted(
                as_of_to_common_prefix.keys()
            )[-1]
            logger.debug(
                'Load type is full, so returning most recent '
                f"as_of_ common prefix: '{most_recent_as_of}'"
            )
            return as_of_to_common_prefix[most_recent_as_of]
        else:
            # Edge case - the `load_type` is `Full Load` but
            # there are zero as_of_ keys, so return and empty list. The
            # code calling this filter will have to return a user
            # friendly error message about there being no files in the
            # dataset.
            # Alternatively, the cache might not be up to date causing
            # the most recent as_of_ to not be in the list.
            logger.warning(
                'Zero as_of_ keys in the partitions but load type is '
                'full. We may have returned some paths that did not '
                'have an as_of_ key in their common_prefix. Alternatively, '
                'we may need to wait for the cache to update.'
            )
            return []



    @staticmethod
    def filter_hidden_prefixes_by_dataset_content_and_load_type(
        content_type: str,
        load_type: str,
        prefix: str,
        unique_common_prefixes: Iterator[str]
    ) -> List[str]:

        unique_non_hidden_common_prefixes = set([])

        # Must be a iter type as we iterate more than once.
        non_hidden_common_prefixes: List[str] = sorted(set(
            [
                unique_common_prefix for unique_common_prefix in unique_common_prefixes
                if not is_hidden_based_on_path_and_content(_shorten(unique_common_prefix, prefix), content_type)
            ]
       ))
        # already clear of as_of= latest at this point

        logger.debug(
            f"load_type from Catalogue: '{load_type}'",
            extra={
                'load_type': load_type,
            }
        )

        if load_type == 'Full Load':
            # For a Structured dataset, we expect the first common prefix on S3
            # to be an `as_of` representing when the data was published.
            #
            # If the dataset's load type is "Full Load"
            # then return only data from the most recent `as_of` common prefix.
            unique_non_hidden_common_prefixes.update(S3DatasetHandler.get_most_recent_common_prefix(
                common_prefixes=iter(non_hidden_common_prefixes)
            ))
        else:
            # Else (if the dataset's load type is "Incremental Load" or None),
            # then we return data from all as_of common prefixes.

            for common_prefix in non_hidden_common_prefixes:
                # We are safe to do an S3 list efficiency of asking for `as_of` instead
                # of asking for every `as_of_*=value`. This reduces the number of calls
                # to S3 LIST.
                #
                # We do not need the entire path of each non-hidden
                # common prefix. We only need to know the prefix and the
                # partition name of the common prefix (if exists) knowing
                # it will not be a `.` or `_`. We do not want to keep the
                # unique value of the partition as that will cause use to
                # make an S3 call for each unique partition.
                # e.g. for common_prefix
                #     abc/as_of_date=2020-01-01
                # we only store
                #     abc/as_of_date
                # and so all the as_of_dates are collected in a single
                # S3 list operation with prefix starting abc/as_of_date.
                if '=' in common_prefix:
                    prefix_plus_partition, _ = common_prefix.split('=', 1)
                    unique_non_hidden_common_prefixes.add(prefix_plus_partition)
                else:
                    logger.warning("This dataset's top-level on S3 is not partitioned")
                    # Edge case where the top-level on S3 is not partitioned, so add the whole
                    # common prefix.
                    unique_non_hidden_common_prefixes.add(common_prefix)

        logger.debug(
            f'length of non_hidden_common_prefixes: '
            f'{len(non_hidden_common_prefixes)}'
            f'\nlength of unique_non_hidden_common_prefixes: '
            f'{len(unique_non_hidden_common_prefixes)}'
        )

        return sorted(set(unique_non_hidden_common_prefixes))

    def get_assumed_role_session(self, dataset: Dict):
        s2 = time.time()
        if 'location' in dataset['attributes'] and 's3' in dataset['attributes']['location']:
            logging.debug("Assuming: "+dataset['attributes']['location']['s3']['aws_role_arn'])
            session = self.create_session(
                role_arn=dataset['attributes']['location']['s3']['aws_role_arn']
            )
            s3 = time.time()
            logger.debug(f"STS {s3 - s2}")
            return session
        else:
            raise UnableToAccessDataset(
                details=(
                    'You do not have access to this dataset. You must '
                    'request access to the dataset\'s package.'
                )
            )

    def __repr__(self):
        return "S3DatasetHandler"


import os
from typing import Optional

import requests
from injector import singleton as singleton_scope, provider
from requests import Session

from ..authentication.auth import _sam_app_flow
from ..context.environments import _Environment
from ..handlers.analytics_handler import AnalyticsHandler, AnalyticsSender
from ..providers.base_provider import BaseModule


class AnalyticsDependencyProvider(BaseModule):
    """
    This is currently used by the s3-proxy for analytics. If you want to use it for other services then you will need
    to override the `app_name` in the config dict.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        config = {
            'version': 'latest',
            'app_name': 's3proxy',
            'consumption_env': os.environ.get('CONSUMPTION_API_URL'),
            'strict': False,
        }

        catalogue_url = os.environ.get('CATALOGUE_API_URL')  # 'https://catalogue-qa.udpmarkit.net'
        environment = _Environment(catalogue_url)

        try:
            decoded_token = _sam_app_flow(
                environment,
                os.environ.get('DLI_ACCESS_KEY_ID'),
                os.environ.get('DLI_SECRET_ACCESS_KEY')
            )
        except Exception:
            decoded_token = None

        def new_session() -> Optional[requests.Session]:
            session = Session()
            session.headers.update({'Authorization': f"Bearer {decoded_token}"})
            return session

        def no_session() -> Optional[requests.Session]:
            return None

        # if DLI_ACCESS_KEY_ID and DLI_SECRET_ACCESS_KEY is
        # not valid we cannot retrieve a JWT from SAM.
        if decoded_token is not None:
            session_func = new_session
        else:
            session_func = no_session

        rest_analytics_sender = AnalyticsSender(config, session_func)
        self._analytics_handler = AnalyticsHandler(rest_analytics_sender)

    @singleton_scope
    @provider
    def analytics_handler(self) -> AnalyticsHandler:
        return self._analytics_handler



#
# Copyright (C) 2019 IHS Markit.
# All Rights Reserved
#

from ..exceptions import ServiceException


class NoFilesInDataset(ServiceException):
    pass


class UnsupportedDatasetFormat(ServiceException):
    pass


class NonParquetFilesInDataset(ServiceException):
    pass


class NonCsvFilesInDataset(ServiceException):
    pass



#
# Copyright (C) 2019 IHS Markit.
# All Rights Reserved
#

import yaml
import logging
import logging.config
import os
from flask import current_app, request, g
from logging import Filter
from ddtrace.helpers import get_correlation_ids


suffixes = ['B', 'KB', 'MB', 'GB', 'TB', 'PB']


def human_memory_size(nbytes):
    i = 0
    while nbytes >= 1024 and i < len(suffixes)-1:
        nbytes /= 1024.
        i += 1
    f = ('%.2f' % nbytes).rstrip('0').rstrip('.')
    return '%s %s' % (f, suffixes[i])


_UNSAFE_HEADERS = ['Authorization']


class DDTraceFilter(Filter):

    def filter(self, record):
        trace_id, span_id = get_correlation_ids()
        if trace_id and span_id:
            record.__dict__['dd.trace_id'] = str(trace_id)
            record.__dict__['dd.span_id'] = str(span_id)
        else:
            record.__dict__['dd.trace_id'] = 0
            record.__dict__['dd.span_id'] = 0

        if request and current_app:
            if record.levelno >= logging.ERROR:
                record.__dict__.update(get_request_data())
                record.__dict__.update(get_safe_headers())

            request_id = getattr(g, 'request_id', None)

            if request_id:
                record.__dict__['request_id'] = request_id

        return True


class ProcessMemoryInfoTraceFilter(Filter):

    def filter(self, record):
        # NOTE keep this import inside the function. We do not want to
        # reference psutil to make sure there is not a side-effect of starting
        # any profiling when it is disabled.
        import psutil
        current_thread = psutil.Process(os.getpid())
        # RSS: resident set size, the portion of memory occupied by
        # a process that is held in main memory
        record.thread_memory_usage = current_thread.memory_info().rss
        # Easier to read (harder to filter by)
        record.thread_memory_usage_human_readable = human_memory_size(
            record.thread_memory_usage
        )
        # NOTE thread_id is usually added to the logger via the formatter
        # but I want to do it all in the same place.
        record.process_id = current_thread.pid
        return True


# we need this due to the fact that some libraries - i.e. numpy,
# return locals with dict containing types as keys.
# `logging` cannot handle that and raises TypeError,
# that is handled internally, resulting in badly formatted logs.
class UnparsableNamespaceFilter(Filter):

    def filter(self, record):
        def _filter_out_unparsable_params(params):
            if not isinstance(params, dict):
                return params

            return {
                k: _filter_out_unparsable_params(v) for k, v in params.items()
                if isinstance(k, (str, int, float, bool)) or k is None
            }

        if hasattr(record, 'locals'):
            record.locals = _filter_out_unparsable_params(record.locals)
        return True


def get_safe_headers():
    # Remove headers that would compromise the user
    safe_headers = dict(request.headers)

    for header in _UNSAFE_HEADERS:
        if header in safe_headers:
            # Don't print entire secret. Allows us to filter by it
            safe_headers[header] = safe_headers[header][:15] + '****'

    return {
        'headers': safe_headers,
    }


def get_request_data():
    return {
        'method': request.method,
        'path': request.path,
        'host': request.host,
        # Request ID for logging between other services.
        'request_id': getattr(g, 'request_id', None)
    }


def setup_logging(
    config_path, level_override=logging.INFO,
    simple_formatter=False,
    pretty_json=False
):
    with open(config_path, 'r') as f:
        config = yaml.load(f, Loader=yaml.SafeLoader)

    config['loggers']['']['level'] = level_override
    config['loggers']['api']['level'] = level_override
    config['loggers']['lib']['level'] = level_override
    config['loggers']['runner']['level'] = level_override
    config['loggers']['webapp']['level'] = level_override

    if simple_formatter:
        config['handlers']['info_handler']['formatter'] = 'simple'

    if pretty_json:
        config['formatters']['json_verbose']['json_indent'] = 4

    logging.config.dictConfig(config)
    logging.captureWarnings(True)



import argparse
import os
import uuid
import random
import requests
from random import choice
from string import digits
import subprocess


ACCOUNTS_URL = None
INTERFACE_URL = None
LOCALSTACK_URL = None
DEBUG = os.environ.get("DEBUG")


def debug_print(to_print):
    if DEBUG == '2':
        print(to_print)


def get_data_org(org_shortcode):
    return {
        "data": {
            "attributes": {
                "name": org_shortcode,
                "description": "string",
                "external_company_id": str(uuid.uuid4()),
                "is_visible": "true",
                "domains": [
                    {"domain_name": f"{org_shortcode}.com"}
                ],
                "short_code": org_shortcode
            }
        }
    }


def get_data_sub_org(org_shortcode, org_id):
    return {
        "data": {
            "attributes": {
                "name": f"{org_shortcode}_sub",
                "description": "string",
                "organisation_id": org_id,
            }
        }
    }


def get_data_aws_accounts(aws_account_id, org_id):
    return {
        "data": {
            "attributes": {
                "vendor_aws_account_id": aws_account_id,
                "vendor_aws_account_name": aws_account_id,
                "aws_region": "eu-west-1",
                "aws_role_arn": "arn:aws:iam::000000000001:user/root",
                "organisation_id": org_id
            }
        }
    }


def get_data_aws_account_role(account_id):
    return {
        "owner_type": "aws-accounts",
        "owner_id": account_id,
        "role": "aws-account-user"
    }


def get_data_sub_org_role(suborg_id, role):
    return {
        "owner_type": "suborganisations",
        "owner_id": suborg_id,
        "role": role
    }


def get_data_packages(package_name, suborg_id):
    return {
        "data": {
            "attributes": {
                "name": package_name,
                "description": "This is a package created for testing S3Proxy.",
                "keywords": ["testing"],
                "topic": "Automotive",
                "access": "Unrestricted",
                "is_internal_within_organisation": "false",
                "terms_and_conditions": "There are no terms and conditions",
                "publisher": "b30a3e2c-ad7e-11eb-b02d-377b10141544",
                "suborganisation_id": suborg_id
            }
        }
    }


def get_data_datasets(sql_enabled, dataset_shortcode, PACKAGE_ID, load_type, AWS_ACCOUNT_ID, bucket, prefix):
    return {
        "data": {
            "attributes": {
                "sql_enabled": sql_enabled,
                "earliest_date_field": "2020",
                "data_preview_type": "NONE",
                "publishing_frequency": "Hourly",
                "naming_convention": "test",
                "documentation": "test",
                "description": "test",
                "content_type": "Structured",
                "data_format": "PARQUET",
                "short_code": dataset_shortcode,
                "package_id": PACKAGE_ID,
                "load_type": load_type,
                "details_visible_to_lite_users": "false",
                "has_datafile_monitoring": "false",
                "taxonomy": ["Forecasting"],
                "location": {
                    "s3": {
                        "type": "S3",
                        "owner": {"aws_account_id": AWS_ACCOUNT_ID},
                        "bucket": bucket,
                        "prefix": prefix
                    }
                },
                "keywords": ["test"],
                "name": dataset_shortcode
            }
        }
    }


def uuid_mod_string():
    s = str(uuid.uuid4())
    s = s.replace("-","")
    num_upper = random.randint(0, 10)
    for i in range(num_upper):
        idx = random.randint(0, 31)
        if not isinstance(s[idx], int):
            s = s[:idx] + s[idx].upper() + s[idx:]
    return s


def get_random_digits():
    return ''.join(choice(digits) for i in range(5))


def check_response_code(fn):
    def wrapper(*args, **kwargs):
        output = fn(*args, **kwargs)
        if output.status_code != 200 and output.status_code != 201:
            raise Exception(f"[CATALOGUE SETUP] [{fn.__name__}] Failure, non 200 response code: {output.json()}, Args: {args}, Kwargs: {kwargs}")
        return output
    return wrapper


@check_response_code
def make_catalogue_get(url, endpoint, headers, data=None):
    output = requests.get(f"http://{url}/{endpoint}", headers=headers, data=data)  # TODO
    debug_print(f"{output}")
    return output


@check_response_code
def make_catalogue_post(url, endpoint, headers, data=None):
    output = requests.post(f"http://{url}/{endpoint}", headers=headers, json=data)
    debug_print(f"{output}")
    return output


def get_random_for_empty(
        jwt_token, org_shortcode, package_name, dataset_shortcode, load_type,
        bucket, prefix, local_data, sql_enabled
):
    jwt_token = jwt_token
    if jwt_token == '':
        raise Exception("[CATALOGUE SETUP] Please pass a test JWT")
    load_type = 'Incremental Load' if load_type is None else 'Full Load'
    org_shortcode = uuid_mod_string() if org_shortcode is None else org_shortcode
    package_name = uuid_mod_string() if package_name is None else package_name
    dataset_shortcode = uuid_mod_string() if dataset_shortcode is None else dataset_shortcode
    bucket = uuid_mod_string().lower() if bucket is None else bucket
    prefix = uuid_mod_string() if prefix is None else prefix
    sql_enabled = 'false' if sql_enabled is None else sql_enabled
    local_data = 'tools/tests/data/integration_dataset/' if local_data is None else local_data
    return package_name, dataset_shortcode, org_shortcode, load_type, bucket, prefix, local_data, sql_enabled


def setup_named_new_dataset_in_named_new_package_in_named_new_org(args=None):
    jwt_token = args.jwt_token
    if jwt_token == '':
        raise Exception("[CATALOGUE SETUP] Please pass a test JWT")
    load_type = 'Incremental Load' if args.load_type is None else 'Full Load'
    org_shortcode = uuid_mod_string() if args.org_shortcode is None else args.org_shortcode
    package_name =uuid_mod_string() if args.package_name is None else args.package_name
    dataset_shortcode = uuid_mod_string() if args.dataset_shortcode is None else args.dataset_shortcode
    bucket = uuid_mod_string().lower() if args.bucket is None else args.bucket
    prefix = uuid_mod_string() if args.prefix is None else args.prefix
    sql_enabled = 'false' if args.sql_enabled is None else args.sql_enabled
    local_data = 'tools/tests/data/integration_dataset/' if args.local_data is None else args.local_data

    setup_new_dataset_in_new_package_in_new_org(jwt_token, load_type, org_shortcode, package_name, dataset_shortcode, bucket, prefix, sql_enabled, local_data)

    # print(org_shortcode, ORG_ID, SUBORG_ID,
    #       package_name, PACKAGE_ID,
    #       dataset_shortcode, dataset_id,
    #       bucket, prefix, AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY)


def setup_new_dataset(
    jwt_token, bucket, prefix, local_data, sql_enabled, dataset_shortcode, PACKAGE_ID, load_type, AWS_ACCOUNT_ID
):
    global ACCOUNTS_URL
    global INTERFACE_URL
    global LOCALSTACK_URL

    headers = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {jwt_token}"
    }

    # upload data to localstack ----------------------------------------------------------------------------
    def call_subprocess(cmd):
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        out, err = proc.communicate()
        if len(err) > 1:
            raise Exception(f"[CATALOGUE SETUP] Failed for AWS command: {cmd}, ERR: {err}")
        debug_print(f"AWS CMD: {cmd}, OUT: {out}, ERR: {err}")

    aws_command_1 = f"aws s3api create-bucket --bucket \"preview-bucket\" --endpoint-url=\"http://{LOCALSTACK_URL}\" --region \"us-east-1\""
    aws_command_2 = f"aws s3api create-bucket --bucket \"{bucket}\" --endpoint-url=\"http://{LOCALSTACK_URL}\" --region \"us-east-1\""
    aws_command_3 = f"aws s3 cp --recursive --endpoint-url=\"http://{LOCALSTACK_URL}\" {local_data} \"s3://{bucket}/{prefix}/\" --region \"us-east-1\""
    cmds = [aws_command_1, aws_command_2, aws_command_3]
    for cmd in cmds:
        call_subprocess(cmd)

    # get datasets ----------------------------------------------------------------------------
    debug_print(f"[CATALOGUE SETUP] [setup_new_dataset_in_new_package_in_new_org] get datasets...")
    datasets_data = get_data_datasets(sql_enabled, dataset_shortcode, PACKAGE_ID, load_type, AWS_ACCOUNT_ID, bucket, prefix)
    outp = make_catalogue_post(INTERFACE_URL, "__api_v2/datasets/", headers, datasets_data)
    datasets = outp.json()
    print(datasets)
    dataset_id = datasets['data']['id']
    return dataset_id


def setup_new_package(jwt_token, package_name, SUBORG_ID):
    global ACCOUNTS_URL
    global INTERFACE_URL
    global LOCALSTACK_URL
    headers = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {jwt_token}"
    }
    # get packages ----------------------------------------------------------------------------
    debug_print(f"[CATALOGUE SETUP] [setup_new_dataset_in_new_package_in_new_org] get packages...")
    package_data = get_data_packages(package_name, SUBORG_ID)
    packages = make_catalogue_post(INTERFACE_URL, "__api_v2/packages/", headers, package_data)
    PACKAGE_ID = packages.json()['data']['id']
    return PACKAGE_ID


def setup_new_dataset_in_new_package(
    INTERFACE=None, LOCALSTACK=None, jwt_token=None, SUBORG_ID=None, ORG_ID=None, AWS_ACCOUNT_ID=None,
    package_name=None, dataset_shortcode=None, load_type=None, bucket=None, prefix=None,
    local_data=None, sql_enabled=None, org_shortcode=None
):
    headers = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {jwt_token}"
    }

    try:
        global ACCOUNTS_URL
        global INTERFACE_URL
        global LOCALSTACK_URL
        INTERFACE_URL = INTERFACE
        ACCOUNTS_URL = INTERFACE_URL
        LOCALSTACK_URL = LOCALSTACK

        if AWS_ACCOUNT_ID is None:
            _, AWS_ACCOUNT_ID, _ = setup_aws_account_for_user(ORG_ID, headers)

        package_name, dataset_shortcode, org_shortcode, load_type, bucket, prefix, local_data, sql_enabled = get_random_for_empty(
            jwt_token, org_shortcode, package_name, dataset_shortcode, load_type,
            bucket, prefix, local_data, sql_enabled
        )
        PACKAGE_ID = setup_new_package(jwt_token, package_name, SUBORG_ID)
        dataset_id = setup_new_dataset(jwt_token, bucket, prefix, local_data, sql_enabled, dataset_shortcode, PACKAGE_ID, load_type, AWS_ACCOUNT_ID)
        return PACKAGE_ID, dataset_id

    except Exception as e:
        print(f"[CATALOGUE SETUP] Setup has failed: {e}")


def setup_new_dataset_in_new_package_in_new_org(
    jwt_token, load_type, org_shortcode, package_name, dataset_shortcode, bucket, prefix, sql_enabled, local_data
):
    headers = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {jwt_token}"
    }
    try:
        debug_print(f"[CATALOGUE SETUP] [setup_new_dataset_in_new_package_in_new_org] setup_org...")
        ORG_ID, SUBORG_ID = setup_org(org_shortcode, headers)

        debug_print(f"[CATALOGUE SETUP] [setup_new_dataset_in_new_package_in_new_org] setup_user_package_creation...")
        (NEW_GROUP_ID, SUBORG_ID, USER_ID, AWS_ACCOUNT_ID, ACCOUNT_ID,
         ADDITIONAL_ORGS_NEW, ADDITIONAL_ORGS_CURRENT) = setup_user_package_creation(ORG_ID, SUBORG_ID, headers)
        _ = setup_new_dataset_in_new_package(jwt_token, ORG_ID, SUBORG_ID, AWS_ACCOUNT_ID, package_name,org_shortcode, dataset_shortcode, load_type, bucket, prefix, local_data, sql_enabled)

    except Exception as e:
        print(f"[CATALOGUE SETUP] Setup has failed: {e}")


def setup_org(org_shortcode, headers):
    debug_print(f"[CATALOGUE SETUP] [setup_org] organisations")
    org_data = get_data_org(org_shortcode)
    outp = make_catalogue_post(ACCOUNTS_URL, "__api_v2/organisations", headers, org_data)
    json_output = outp.json()
    debug_print(json_output)
    org_id = json_output["data"]["id"]
    debug_print(f"ORG_ID: {org_id}")

    debug_print(f"[CATALOGUE SETUP] [setup_org] suborganisations")
    suborg_data = get_data_sub_org(org_shortcode, org_id)
    outp = make_catalogue_post(ACCOUNTS_URL, "__api_v2/suborganisations", headers, suborg_data)
    json_output = outp.json()
    debug_print(json_output)
    suborg_id = json_output["data"]["id"]
    debug_print(f"SUB_ORG_ID: {suborg_id}")
    return org_id, suborg_id


def setup_aws_account_for_user(ORG_ID, headers):

    # register aws account for users ----------------------------------------------------------------------------
    debug_print(f"[CATALOGUE SETUP] [setup_aws_account_for_user] register aws account for users")
    AWS_ACCOUNT_ID = get_random_digits()
    aws_accounts_data = get_data_aws_accounts(AWS_ACCOUNT_ID, ORG_ID)
    outp = make_catalogue_post(INTERFACE_URL, "__api_v2/aws_accounts", headers, aws_accounts_data)
    json_output = outp.json()
    debug_print(json_output)
    ACCOUNT_ID = json_output['data']['id']

    # TODO Disable after Catalogue DL-7224
    # get our user group --------------------------------------------------------------------------------------------
    debug_print(f"[CATALOGUE SETUP] [setup_aws_account_for_user] get our user group")
    outp = make_catalogue_get(ACCOUNTS_URL, "__api_v2/user_groups/", headers)  # NOTE: this is a GET
    json_output = outp.json()
    debug_print(json_output)
    group_id = json_output['data'][0]['id']
    NEW_GROUP_ID = group_id

    # add the aws_account to our user's group -----------------------------------------------------------------------
    debug_print(f"[CATALOGUE SETUP] [setup_aws_account_for_user] add the aws_account to our user's group")
    role_data = get_data_aws_account_role(ACCOUNT_ID)
    outp = make_catalogue_post(ACCOUNTS_URL, f"__api_v2/user_groups/{NEW_GROUP_ID}/roles", headers, role_data)
    json_output = outp.json()
    debug_print(json_output)

    return NEW_GROUP_ID, AWS_ACCOUNT_ID, ACCOUNT_ID


def setup_user_package_creation(org_id, SUBORG_ID, headers):

    debug_print(f"[CATALOGUE SETUP] [setup_user_package_creation] [setup_aws_account_for_user]")
    NEW_GROUP_ID, AWS_ACCOUNT_ID, ACCOUNT_ID = setup_aws_account_for_user(org_id, headers)

    # set permissions on suborg for our user now we have the group id -----------------------------------------------
    debug_print(f"[CATALOGUE SETUP] [setup_user_package_creation] set permissions on suborg for our user now we have the group id")
    role = 'data-steward'
    role_data = get_data_sub_org_role(SUBORG_ID, role)
    outp = make_catalogue_post(ACCOUNTS_URL, f"__api_v2/user_groups/{NEW_GROUP_ID}/roles", headers, role_data)
    json_output = outp.json()
    debug_print(json_output)

    # set permissions on suborganisations with role technical-support for our user ----------------------------------
    debug_print(f"[CATALOGUE SETUP] [setup_user_package_creation] set permissions on suborganisations with role technical-support for our user")
    role = 'technical-support'
    role_data = get_data_sub_org_role(SUBORG_ID, role)
    outp = make_catalogue_post(ACCOUNTS_URL, f"__api_v2/user_groups/{NEW_GROUP_ID}/roles", headers, role_data)
    json_output = outp.json()
    debug_print(json_output)

    # set permissions on suborganisations with role access-manager for our user ------------------------------------
    debug_print(f"[CATALOGUE SETUP] [setup_user_package_creation] set permissions on suborganisations with role access-manager for our user")
    role = 'access-manager'
    role_data = get_data_sub_org_role(SUBORG_ID, role)
    outp = make_catalogue_post(ACCOUNTS_URL, f"__api_v2/user_groups/{NEW_GROUP_ID}/roles", headers, role_data)
    json_output = outp.json()
    debug_print(json_output)

    # set permissions on suborganisations with role package-creator for our user ----------------------------------
    debug_print(f"[CATALOGUE SETUP] [setup_user_package_creation] set permissions on suborganisations with role package-creator for our user")
    role = 'package-creator'
    role_data = get_data_sub_org_role(SUBORG_ID, role)
    outp = make_catalogue_post(ACCOUNTS_URL, f"__api_v2/user_groups/{NEW_GROUP_ID}/roles", headers, role_data)
    json_output = outp.json()
    debug_print(json_output)

    # TODO label in bash to uncomment code for ADDITIONAL_ORGS_NEW, ADDITIONAL_ORGS_CURRENT
    USER_ID=None
    ADDITIONAL_ORGS_NEW=None
    ADDITIONAL_ORGS_CURRENT=None

    return NEW_GROUP_ID, SUBORG_ID, USER_ID, AWS_ACCOUNT_ID, ACCOUNT_ID, ADDITIONAL_ORGS_NEW, ADDITIONAL_ORGS_CURRENT


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("--accounts", "-a", type=str, required=True)
    parser.add_argument("--interface", "-i", type=str, required=True)

    parser.add_argument("--localstack", "-l", type=str, required=True)

    parser.add_argument("--jwt_token", "-j", type=str, required=False)
    parser.add_argument("--load_type", "-t", type=str, required=False)
    parser.add_argument("--org_shortcode", "-o", type=str, required=False)
    parser.add_argument("--package_name", "-p", type=str, required=False)
    parser.add_argument("--dataset_shortcode", "-s", type=str, required=False)
    parser.add_argument("--bucket", "-b", type=str, required=False)
    parser.add_argument("--prefix", "-x", type=str, required=False)
    parser.add_argument("--sql_enabled", "-e", type=str, required=False)
    parser.add_argument("--local_data", "-d", type=str, required=False)
    parser.add_argument("--mode", "-m", type=str, required=True)

    args = parser.parse_args()
    ACCOUNTS_URL = args.interface #args.accounts
    LOCALSTACK_URL = args.localstack
    INTERFACE_URL = args.interface  # to be args.accounts
    ACCOUNTS = INTERFACE_URL
    INTERFACE = INTERFACE_URL
    LOCALSTACK = LOCALSTACK_URL

    DEBUG = os.environ.get("DEBUG", "2")
    if not ACCOUNTS_URL or not INTERFACE_URL or not LOCALSTACK_URL:
        raise Exception(f"Environment variables are not set: "
                        f"ACCOUNTS_URL:{ACCOUNTS_URL}"
                        f"INTERFACE_URL:{INTERFACE_URL}"
                        f"LOCALSTACK_URL:{LOCALSTACK_URL}"
                        )
    # setup_named_new_dataset_in_named_new_package_in_named_new_org(args)
    pass


from typing import List, Tuple, Dict


class DLCFilterAdapter(List[Tuple[str, str, str]]):
    def to_params(self) -> Dict[str, str]:
        result = {}

        for idx, param in enumerate(self):
            field, operator, value = param
            result.update({
                f'filter[{idx}][field]': field,
                f'filter[{idx}][operator]': operator,
                f'filter[{idx}][value]': value,
            })

        return result



#
# Copyright (C) 2019 IHS Markit.
# All Rights Reserved
#
import logging
from contextlib import contextmanager
from pyarrow import ArrowException
from pyarrow.lib import ArrowNotImplementedError

from ...arrow_service.exceptions import NonCsvFilesInDataset, NonParquetFilesInDataset
from ....services.exceptions import ServiceException, SchemaMismatchException

logger = logging.getLogger(__name__)

# TODO: this seems redundant now we have utils.stream_error_chunker wrapping
# all of our streams. This is called in handle()


@contextmanager
def handle_arrow_stream_errors(writer):
    message = 'Unknown Exception in arrow stream. ' \
              'This could be a problem with the file you are trying ' \
              'to read. Please report the problem to the datalake ' \
              'platform team at DataLake-Support_L1@ihsmarkit.com. ' \
              'Please include the full error message including ' \
              'the request_id.'

    try:
        yield writer
    except NonParquetFilesInDataset:
        logger.warning('NonParquetFilesInDataset')
        raise
    except NonCsvFilesInDataset:
        logger.warning('NonCsvFilesInDataset')
        raise
    except ArrowNotImplementedError as e:
        logger.exception('Arrow not implemented error')
        if ('structs' and 'Parquet files') in e.args[0]:
            raise ServiceException(
                details='The dataset contains a Parquet datafile with '
                        'nested structs, reading from it is not yet '
                        'supported in this version of pyarrow.'
            ) from e
        raise ServiceException(
            details='The dataset contains a datafile '
                    'that cannot be read by pyarrow.'
            ) from e

    # TODO scott: can this ever be reached if we catch it in the handler?
    except ArrowException as e:
        # This will generally happen outside the
        # request context so there's nothing catching
        # the exception and re-logging it in a structured form.

        # This is log level warning rather than error because it can be
        # caused by a csv in a parquet dataset.
        logger.warning(
            'ArrowException while iterating over chunk',
            exc_info=e,
        )
        raise ServiceException(message) from e
    except SchemaMismatchException:
        logger.warning('SchemaMismatchException')
        raise
    except Exception as e:
        logger.exception('Unknown Exception in arrow stream.')
        raise ServiceException(message) from e
    finally:
        logger.debug('Closing writer')
        writer.close()



import operator
from typing import List, Dict
from flask import current_app
try:
    # these are only present for services, not for the SDK, and we do not wish to impose these in requirements.txt
    from injector import inject
    from pymemcache import HashClient
except ModuleNotFoundError:
    # set up an Identity decorator for @inject
    def inject(function): # type: ignore
        def wrapper(*args, **kwargs):
            return function(*args, *kwargs)
        return wrapper
except:
    pass

import logging

from ..providers.oidc_provider import JWT
from ..services.abstract_backend_handler_service import AbstractBackendDatasetHandler
from ..services.view_exceptions import NotFound
from ..services.exceptions import ServiceException, InterfaceResourceUnavailable
from ..providers.memcache_provider import memcached_method
from ..services.dlc_service import DlcService
from ..services.identity_service import IdentityService

logger = logging.getLogger(__name__)


class CachedFetcherService:
    @inject
    def __init__(
        self,
        dlc_service: DlcService,
        identity_service: IdentityService,
        hash_client: 'HashClient'
    ):
        logger.debug('Init CachedFetcherService')
        self.dlc_service = dlc_service
        self.identity_service = identity_service
        self.hash_client = hash_client

    @memcached_method(cache_getter=operator.attrgetter('hash_client'), )
    def list_objects_in_dataset(self, organisation_shortcode: str, args, method_name: str,
                                handler: AbstractBackendDatasetHandler) -> Dict:
        jwt = current_app.injector.get(JWT)
        datasets: List[Dict] = self.dlc_get_accessible_datasets(
            jwt=jwt,
            organisation_shortcode=organisation_shortcode,
            key=args['Prefix']
        )
        return handler.list(datasets, organisation_shortcode, args, method_name)

    def dlc_get_accessible_datasets(
            self, jwt: str, organisation_shortcode: str, key: str = '',
    ) -> List[Dict]:
        """
        Only returns datasets the user can access (by checking visible
        organisations and the dataset's `location` field).

        Raises if no matching visible organisation for the user.

        SECURITY Do not cache the result of this function, there is too much logic here and
        not all of the parameters (such as `key`) should be in all of the caches.
        Instead each call this function makes is cached.

        :param jwt: SECURITY. Used to make caching specific to the user. Must be present.
        :param organisation_shortcode:
        :param key: If empty we need to list all the datasets.
        If partial with need to match the visible datasets starting with
        the short code.
        If full we need to list files inside dataset.
        If full with path we need to list files inside plus that path.
        :return: List of datasets the user can access.
        """

        # Raises if no matching visible organisation for the user.
        # self.get_organisation(
        #     jwt=jwt,  # Needed so that cache value is secured against the JWT.
        #     organisation_shortcode=organisation_shortcode,
        # )

        # Now get the dataset.

        # Some example keys
        # ''                          - Empty
        # 'data_s'                    - Possibly Incomplete
        # 'data_set1'                 - Complete dataset
        #                                 but search for other paths
        # 'data_set1/'                - Complete dataset
        # 'data_set1/path/to/file'    - Dataset with path

        # If empty we need to list all the datasets.
        # If partial with need to match the visible datasets starting with
        # the short code.
        # If full we need to list files inside dataset.
        # If full with path we need to list files inside plus that path.

        # Sanity check.
        if key and key.lower().startswith('s3:'):
            raise ServiceException(
                details='You have asked for a key that begins with `s3:`, which is an '
                        'absolute path. Please replace this with key that is a relative path. '
                        'The key should with the dataset short code '
                        'e.g. key=example-dataset-short-code   or  '
                        'key=example-dataset-short-code/as_of_date=2020-01-01 '
            )

        # noinspection PyBroadException
        try:
            func = self.dlc_service.get_datasets_in_organisation
            args = {
                "jwt": jwt,
                "organisation_short_code": organisation_shortcode,
            }

            if not key:
                # Empty key is the simplest case - we just get given /ORGANISATION
                return func(**args)

            dataset_shortcode, has_prefix_following, _ = key.partition('/')
            if not has_prefix_following:
                # We get given /ORGANISATION/DATASET
                return func(**args, shortcode_startswith=dataset_shortcode)
            else:
                # We get given /ORGANISATION/DATASET/...
                dataset = [self.dlc_service.get_dataset_for_shortcode(**args, dataset_shortcode=dataset_shortcode)]
                return dataset

        except InterfaceResourceUnavailable:
            # This means there's no match to return
            logger.debug(
                'No orgs found for short_code.',
                extra={'organisation_shortcode': organisation_shortcode}
            )
            return []
        except Exception:
            logger.exception('Unhandled exception in dlc_get_accessible_datasets')
            return []

    @memcached_method(cache_getter=operator.attrgetter('hash_client'), )
    def dlc_get_organisation(self, jwt: str, organisation_shortcode: str) -> dict:
        """
        Raises if:
        organisation_shortcode does not exist
        or the organisation is not visible to the user.

        :param jwt: SECURITY. Used to make caching specific to the user. Must be present.
        :param organisation_shortcode:
        :return: An organisation dict.
        """
        try:
            logger.debug(
                'get_organisation',
                extra={
                    'organisation_shortcode': organisation_shortcode,
                }
            )
            organisations = self.dlc_get_visible_organisations(jwt=jwt)

            for org in organisations['data']:
                if org['attributes']['short_code'].lower() == organisation_shortcode.lower():
                    return org

            logger.warning(
                'Attempt to get non-existing or non-visible org',
                extra={'organisation_shortcode': organisation_shortcode}
            )

            raise NotFound('Organisation not found')
        except Exception as e:
            logger.error(e)
            return {}

    @memcached_method(cache_getter=operator.attrgetter('hash_client'), )
    def dlc_get_visible_organisations(self, jwt: str) -> dict:
        """
        Get visible organisations.

        :param jwt: SECURITY. Used to make caching specific to the user. Must be present,
        otherwise a different user can see the organisations of this user. This happened in a
        red alert where an external QA user was able to see the internal organisation.
        :return: An organisation dict of only the organisations that are visible to this user's
        JWT.
        """
        try:
            return self.identity_service.get_visible_organisations()
        except Exception as e:
            logger.error(e)
            return {}

    @memcached_method(cache_getter=operator.attrgetter('hash_client'), )
    def dlc_me(self, jwt: str) -> dict:
        try:
            return self.identity_service.me()
        except Exception as e:
            logger.error(e)
            return {}



#
# Copyright (C) 2020 IHS Markit.
# All Rights Reserved
#
import os
from collections import OrderedDict
from typing import Dict, Any, Union, List
from functools import partial
from urllib.parse import urljoin

import requests
import logging
from cachetools import TTLCache, cached

from dli.models.dataset_model import DatasetModel
from dli.modules.package_module import PackageModule

from platform_services_lib.lib.services.exceptions import CatalogueEntityNotFoundException
from platform_services_lib.lib.adapters.pagination_adapters import Paginator
from platform_services_lib.lib.context.urls import dataset_urls, identity_urls


class DatasetModule:
    """
    DatasetModule contains functions that operate across datasets

    e.g. searching for a dataset and getting a specific dataset.
    """

    # Caching was added due to a user calling this in a tight loop, causing load on the Catalogue.
    # The default TTL is set to 60 seconds so that users do not have to wait long after changing
    # their visible organisation setting to see the change reflected in the SDK, and at the same
    # time cutting the load on Catalogue.
    # The TTL value come from the environment variable `DLI_CACHE_TTL`. It is overridable for QA
    # testing. Users should not override this value unless you are sure that you are not expecting
    # the visible organisations list to change while running your code.

    trace_logger = logging.getLogger('trace_logger')

    __visible_orgs_ttl_cache: TTLCache = TTLCache(
        maxsize=16,
        ttl=int(os.environ.get("DLI_CACHE_TTL", '60'))
    )

    def __call__(self, search_term=None, only_mine=False, warn=False, page_size=5000) \
            -> Dict[str, Union[DatasetModel, List[DatasetModel]]]:
        """
        Find datasets. Synonymous with .find(...).

        :Example:

            Get all datasets (including those you do not yet have access to):

            .. code:: python

                client.datasets()

        :Example:

            Get **only** datasets you have access to and
            can read:

            .. code:: python

                client.datasets(only_mine=True)

        :Example:

            Search for datasets using `ilike`. This will find fields that are
            similar to value. This is recommended over using `like` as it
            performs case insensitive matching. Example usages `ilike example`.

            Some examples of how the logic matches in the database:

            'example' ILIKE 'example'    true

            'example' ILIKE 'ex%'     true

            'example' ILIKE 'c'      false

            .. code:: python

                # Exact match `example` (case sensitive).
                client.datasets(search_term='name=example')

                # Match `example` (case insensitive).
                client.datasets(search_term='name ilike example')

                # Match `example` (case insensitive) followed by any characters.
                client.datasets(search_term='name ilike example%')

                # Match `example` (case insensitive) in the middle of any characters.
                client.datasets(search_term='name ilike %example%')

        :Example:

            Search for datasets with multiple search terms:

            .. code:: python

                client.datasets(
                    search_term=[
                        'name=example',
                        'manager_id=example'
                    ]
                )

        :Example:

            Search for datasets with name and only_mine:

            .. code:: python

                client.datasets(
                    search_term='name=example',
                    only_mine=True,
                )


        :param search_term: search term for dataset. Can filter using a simple
            query language. E.G "field=value", "field>value", "field ilike value".
            Can pass in multiple values using a list of strings.

        :param bool only_mine: Specify whether to collect datasets only accessible to you (True) or to discover packages
            that you may want to discover but may not have access to (False).

        :param bool warn: Specify whether to display a warning message when results contain UnstructuredDatasets - as the results may consist of both structured and unstructured types which have different methods.

        :param int page_size: Optional. Default 5000. Used to control the size of internal calls from the SDK to the Catalogue.
            This parameter is only exposed to allow performance testing and changes for batch jobs.
            Customers should use the default setting.

        :returns: Ordered dictionary of ids to DatasetModel.
        :rtype: OrderedDict[id: str, DatasetModel]
        """

        filters = {}
        for x in PackageModule._filter_creation(search_term, only_mine):
            filters.update(x)

        p = Paginator(
            dataset_urls.v2_index,
            self._client._DatasetFactory,
            partial(self._client._DatasetFactory._from_v2_response_unsheathed, warn=warn),
            page_size=page_size,
            max_workers=5,
            filters=filters
        )

        xs = [v for v in p]
        results = [(x.short_code, x) for x in xs]
        o = OrderedDict()
        for code, val in results:
            if code in o:
                if type(o[code]) is not list:
                    # if already exists, we need
                    o[code] = [o[code]]
                o[code].append(val)
            else:
                o[code] = val

        return o

    def _visible_orgs(self) -> Dict[Any, Any]:
        """Make a call to the Catalogue."""
        url = urljoin(self._client._environment.catalogue, identity_urls.orgs_visible_to_user)
        self.trace_logger.info(f'visible_orgs: Make real call to Catalogue, url {url}')
        return self._client.session.get(url).json()

    @cached(cache=__visible_orgs_ttl_cache)
    def _visible_orgs_cache_func(self, jwt: str) -> Dict[Any, Any]:
        return self._visible_orgs()

    def find(self, search_term=None, only_mine=False, warn=False, page_size=5000) \
            -> Dict[str, Union[DatasetModel, List[DatasetModel]]]:
        """
        Find datasets matching search terms and whether to only find datasets which you can access.

        :Example:

            Get all datasets (including those you do not yet have access to):

            .. code:: python

                client.datasets.find()

        :Example:

            Get **only** datasets you have access to and
            can read:

            .. code:: python

                client.datasets.find(only_mine=True)

        :Example:

            Search for datasets using `ilike`. This will find fields that are
            similar to value. This is recommended over using `like` as it
            performs case insensitive matching. Example usages `ilike example`.

            Some examples of how the logic matches in the database:

            'example' ILIKE 'example'    true

            'example' ILIKE 'ex%'     true

            'example' ILIKE 'c'      false

            .. code:: python

                # Exact match `example` (case sensitive).
                client.datasets.find(search_term='name=example')

                # Match `example` (case insensitive).
                client.datasets.find(search_term='name ilike example')

                # Match `example` (case insensitive) followed by any characters.
                client.datasets.find(search_term='name ilike example%')

                # Match `example` (case insensitive) in the middle of any characters.
                client.datasets.find(search_term='name ilike %example%')

        :Example:

            Search for datasets with multiple search terms:

            .. code:: python

                client.datasets.find(
                    search_term=[
                        'name like example',
                        'id=0f22cfc8-2a73-11ec-bea8-bf6d5753baf9'
                    ]
                )

        :Example:

            Search for datasets with name and only_mine:

            .. code:: python

                client.datasets.find(
                    search_term='name=example',
                    only_mine=True,
                )


        :param search_term: search term for dataset. Can filter using a simple
            query language. E.G "field=value", "field>value", "field ilike value".
            Can pass in multiple values using a list of strings.

        :param bool only_mine: Specify whether to collect datasets only accessible to you (True) or to discover packages
            that you may want to discover but may not have access to (False).

        :param bool warn: Specify whether to display a warning message when results contain UnstructuredDatasets - as the results may consist of both structured and unstructured types which have different methods.

        :param int page_size: Optional. Default 5000. Used to control the size of internal calls from the SDK to the Catalogue.
            This parameter is only exposed to allow performance testing and changes for batch jobs.
            Customers should use the default setting.

        :returns: Ordered dictionary of ids to DatasetModel.
        :rtype: OrderedDict[id: str, DatasetModel]
        """

        return self.__call__(search_term, only_mine, warn, page_size)

    def get(self, short_code, organisation_short_code=None, page_size=5000) -> DatasetModel:
        """
        Returns a DatasetModel if it exists, and a user has access else None

        :param str short_code: The short code of the dataset to collect

        :param organisation_short_code:  Optional. The short code of the organisation
            for this dataset (useful to specify if you have access to several organisations where the same shortcode exists in several)

        :param int page_size: Optional. Default 5000. Used to control the size
            of internal calls from the SDK to the Catalogue. This parameter is
            only exposed to allow performance testing and changes for batch jobs.
            Customers should use the default setting.

        :returns: Dataset model with matching short code.
        :rtype: DatasetModel
        """

        search_terms = [
            f"short_code={short_code}"
        ]

        if organisation_short_code is not None:
            search_terms.append(
                f"organisation_short_code={organisation_short_code}"
            )

        results: Dict[str, Union[DatasetModel, List[DatasetModel]]] = self._client.datasets(
            search_term=search_terms, only_mine=False, warn=True, page_size=page_size
        )
        res = results.get(short_code)

        if type(res) is list and len(res) > 1:
            # Multiple datasets with same shortcode! This can happen as short codes are only
            # unique within an organisation. We need to call Catalogue to tell the user the
            # visible organisations they have access to so that they can make the call again
            # but with the organisation short code so that only one dataset is returned.

            visible_orgs_response = self._visible_orgs()

            mapping_to_sc = dict(map(
                lambda x: (x["id"],
                           x["attributes"]["short_code"]
                           ),
                visible_orgs_response["data"])
            )

            if os.name != 'nt':
                org_codes=[f"\033[4m\033[92m\033[1morganisation_short_code=" \
                           f"{mapping_to_sc.get(x.organisation_id, x.organisation_id)}" \
                           f"\033[0;0m -> {x.short_code}"
                           for x in res]

                msg = f"\033[96mMultiple datasets have shortcode: \033[1m`{short_code}`\033[0;0m\n\nShortcodes are"\
                      f" only guaranteed to be unique per organisation, and you seem to "\
                      f" have access to multiple organisation's datasets, where there are "\
                      f"datasets bearing the same shortcode."\
                      f"\n"\
                      f"To resolve this, please specify the organisation_short_code in "\
                      f"the call to datasets.get(shortcode, organisation_short_code=ORG CODE)"\
                      f"\n\nThe available (accessible to you) organisations shortcodes "\
                      f"for this dataset short code are:\n" + ("\n".join(org_codes))
            else:
                org_codes = [f"organisation_short_code=" \
                             f"{mapping_to_sc.get(x.organisation_id, x.organisation_id)}" \
                             f" -> {x.short_code}"
                             for x in res]

                msg = f"Multiple datasets have shortcode: `{short_code}`\n\nShortcodes are" \
                      f" only guaranteed to be unique per organisation, and you seem to " \
                      f" have access to multiple organisation's datasets, where there are " \
                      f"datasets bearing the same shortcode." \
                      f"\n" \
                      f"To resolve this, please specify the organisation_short_code in " \
                      f"the call to datasets.get(shortcode, organisation_short_code=ORG CODE)" \
                      f"\n\nThe available (accessible to you) organisations shortcodes " \
                      f"for this dataset short code are:\n" + (
                          "\n".join(org_codes))

            raise Exception(
                msg
            )

            #todo - difficult here, really need to convert id to shortcode for org
            #todo - else isnt that useful a message as user now needs to discover these
        elif res:
            return res
        else:
            if self._client.strict:
                r = requests.Response()
                r.status_code = 404
                r.request = (lambda: True)
                r.request.method = f"No such dataset {short_code}"
                r.request.url = ""
                raise CatalogueEntityNotFoundException(
                    message=f"No such dataset {short_code}",
                    response=r
                )
            else:
                return None



#
# Copyright (C) 2019 IHS Markit.
# All Rights Reserved
#

import dataclasses
from urllib.parse import urljoin
from requests.adapters import DEFAULT_POOLSIZE
from requests import Session
try:
    # these are only present for services, not for the SDK, and we do not wish to impose these in requirements.txt
    from injector import provider, singleton as singleton_scope
except ModuleNotFoundError:
    # set up an Identity decorator for @inject
    def provider(function):  # type: ignore
        def wrapper(*args, **kwargs):
            return function(*args, *kwargs)
        return wrapper

    def singleton_scope(function):  # type: ignore
        def wrapper(*args, **kwargs):
            return function(*args, *kwargs)

        return wrapper

from urllib3 import Retry
import logging

from ..providers.base_provider import BaseModule
from ..adapters.session_adapters import BaseAdapter


@dataclasses.dataclass
class Config:
    api_root: str
    # Defaults taken from the requests library were 10, 10. Under performance
    # tests with 250 concurrent users and 30 pods, 10 was not enough.
    pool_connections: int = DEFAULT_POOLSIZE
    pool_maxsize: int = DEFAULT_POOLSIZE
    # With retries, the total delay of around 32 seconds before a request is
    # deemed "an unrecoverable error"
    # https://urllib3.readthedocs.io/en/latest/reference/
    # urllib3.util.html#urllib3.util.retry.Retry
    max_retries: int = 5


class GAClient(Session):

    logger = logging.getLogger(__name__)

    def __init__(self, config):
        super().__init__()
        self.config = config

        adapter = BaseAdapter(
            pool_connections=self.config.pool_connections,
            pool_maxsize=self.config.pool_maxsize,
            # https://urllib3.readthedocs.io/en/latest/reference/
            # urllib3.util.html#urllib3.util.retry.Retry
            max_retries=Retry(
                total=self.config.max_retries,
                status=self.config.max_retries,
                # List of expected errors
                # https://developers.google.com/analytics/devguides/
                # reporting/core/v4/errors
                status_forcelist=[500, 502, 503, 504],
                method_whitelist=Retry.DEFAULT_METHOD_WHITELIST,
                backoff_factor=1,
            ),
            pool_block=True,
        )

        self.logger.debug(
            f"GAClient session mount prefix '{self.config.api_root}'"
        )
        self.mount(self.config.api_root, adapter)

    def request(self, method, url, *args, **kwargs):
        url = urljoin(self.config.api_root, url)
        return super().request(method, url, *args, **kwargs)


class GADependencyProvider(BaseModule):
    config_class = Config

    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        self.session = GAClient(self.config)

    @singleton_scope
    @provider
    def ga_client(self) -> GAClient:
        return self.session



from pymemcache.client.hash import HashClient
from unittest.mock import MagicMock
import datetime
from freezegun import freeze_time

from ..providers.memcache_provider import MemcachedDependencyProvider

class TestRefreshingHashClient:

    def test__rediscover_and_recreate_client_on_new_config_is_called(self):
        provider = MemcachedDependencyProvider(
            elasticache_config_node_url='test:8000',
            default_ttl=36
            )
        test_client = provider.memcached_client()

        def rediscover_side_effect():
            test_client._last_successful_discovery = datetime.datetime.now()
            test_client._client = MagicMock()

        test_client._rediscover_and_recreate_client_on_new_config = MagicMock(side_effect=rediscover_side_effect)
        test_client.get('h')
        assert test_client._client is not None
        old_magic_mock = test_client._client

        test_client.get('h')
        assert test_client._client is old_magic_mock
        deltatime = test_client._last_successful_discovery + datetime.timedelta(minutes=6)
        with freeze_time(deltatime):
            test_client.get('h')
            assert test_client._client is not old_magic_mock
            assert test_client._client is not None

    def test_should_rediscover(self):
        provider = MemcachedDependencyProvider(
            elasticache_config_node_url='test:8000',
            default_ttl=36
        )
        test_client = provider.memcached_client()
        test_client._last_successful_discovery = datetime.datetime.now()
        assert test_client._should_rediscover() is False

        deltatime = test_client._last_successful_discovery + datetime.timedelta(minutes=6)
        with freeze_time(deltatime):
            assert test_client._should_rediscover() is True

    def test_rediscover_nodes_sets_client_to_hash_client(self):
        provider = MemcachedDependencyProvider(
            elasticache_config_node_url='test:8000',
            default_ttl=36
            )
        test_client = provider.memcached_client()
        test_client._discover_nodes = MagicMock(return_value=['127.0.0.1'])
        test_client._rediscover_and_recreate_client_on_new_config()
        assert test_client._client is not None and isinstance(test_client._client, HashClient)

    def test_client_set_if_no_nodes_returned(self):
        provider = MemcachedDependencyProvider(
            elasticache_config_node_url='test:8000',
            default_ttl=36
        )
        test_client = provider.memcached_client()
        test_client._discover_nodes = MagicMock(return_value=[])
        test_client._rediscover_and_recreate_client_on_new_config()
        assert test_client._client is None



import abc
from typing import List
from cachetools import TTLCache
import logging

try:
    # these are only present for services, not for the SDK, and we do not wish to impose these in requirements.txt
    from pymemcache import HashClient
except ModuleNotFoundError:
    pass

logger = logging.getLogger(__name__)


class AbstractBackendDatasetHandler:
    __metaclass__ = abc.ABCMeta

    def __init__(
            self,
            ttl_cache: TTLCache,
            hash_client: 'HashClient',
    ):
        """
        Constructor.

        :param ttl_cache: tbc
        :param hash_client: Used for caching responses from S3.
        """
        logger.debug('AbstractHandler constructor')
        self.hash_client = hash_client
        self.ttl_cache = ttl_cache

    @abc.abstractmethod
    def handle(self, dataset):
        """
        lists each item in the dataset from the backend storage subsystem in terms of a virtual file handle,
        eg. S3File/ServiceIO
        """
        pass

    @abc.abstractmethod
    def get_assumed_role_session(self, dataset):
        """ assumes any credentials needed by the handler to access the dataset on the backend storage subsystem """
        pass

    @abc.abstractmethod
    def list(self, accessible_datasets: List, organisation_shortcode: str, args, method_name: str):
        """
        listing of the dataset from the backend storage subsystem.
        We do not want to be prescriptive here over the return type, this could be
        - a list of s3 file paths
        - a dictionary of contents, prefixes
        - a list of s3 file paths and sizes

        """
        pass


#
# Copyright (C) 2019 IHS Markit.
# All Rights Reserved
#
import logging
import operator
from typing import Optional, List, Dict
try:
    # these are only present for services, not for the SDK, and we do not wish to impose these in requirements.txt
    from pymemcache.client.hash import HashClient
    from injector import inject
except ModuleNotFoundError:
    # set up an Identity decorator for @inject
    def inject(function): # type: ignore
        def wrapper(*args, **kwargs):
            return function(*args, *kwargs)
        return wrapper

from ..adapters.filter_adapters import DLCFilterAdapter
from ..providers.memcache_provider import memcached_method
from ..providers.catalogue_provider import JWTDlcClient
from ..services.abstract_external_service import ExternalService

logger = logging.getLogger(__name__)


class DlcService(ExternalService):
    """
    Data Lake interface service, for interfacing
    with said interface.

    TODO Remove duplication of code between this and
    JWTDlcClient. The JWTDlcClient should ONLY be responsible
    for authentication and connection pooling. This class
    should be repsonsible for implementation of high level
    interfacing with the interface component.
    """
    @inject
    def __init__(
        self,
        dlc_session: JWTDlcClient,
        hash_client: 'HashClient',
    ):
        logger.debug('Init DlcService')
        self.dlc_session = dlc_session
        self.hash_client = hash_client

    @memcached_method(cache_getter=operator.attrgetter('hash_client'),)
    def get_dataset_for_shortcode(
        self, jwt: str, dataset_shortcode: str,
        organisation_short_code: Optional[str] = None,
    ) -> dict:
        # convert the dataset short code to org
        return self.dlc_session.get(
            f"/__api_v2/by_short_code/dataset/{dataset_shortcode}/",
            params={
                # WARNING! This endpoint as of 30-06-2021 is case sensitive. We have raised
                # ticket DL-7020 for the Catalogue team to make their filter case insensitive.
                'organisation_short_code': organisation_short_code
            },
            hooks=self._make_hook('No dataset for shortcode.')
        ).json()['data']

    @memcached_method(cache_getter=operator.attrgetter('hash_client'),)
    def get_datasets_in_organisation(
        self,
        jwt: str,
        organisation_short_code: str,
        location_type: str = 's3',
        shortcode_startswith: Optional[str] = None,
    ) -> List[dict]:
        """
        Only returns datasets the user can access (by checking the `location`
        field).

        :param jwt: SECURITY. Used to make caching specific to the user. Must be present.
        :param organisation_short_code:
        :param location_type:
        :param shortcode_startswith:
        :return:
        """

        """
        Only returns datasets the user can access (by checking the `location`
        field is not None).

        :param organisation_id:
        :param location_type:
        :param shortcode_startswith:
        :return:
        """

        filters = DLCFilterAdapter([
            # Use ilike to avoid case issues with Catalogue (vs hive style all lowercase).
            # The uniqueness constraint for creating an organisation short code in Catalogue is
            # to check for lowercase e.g. `ihsmarkit`
            # and they store in the database as mixed-case `IHSMarkit`
            # and then the search is against the case sensitive `IHSMarkit`.
            ('organisation_short_code', 'ilike', organisation_short_code),
            ('has_access', 'eq', 'True')
        ])

        if shortcode_startswith:
            filters.append(
                ('short_code', 'startswith', shortcode_startswith)
            )

        params = filters.to_params()
        # Crude substitute for pagination
        params['page_size'] = '5000'

        dataset_response = self.dlc_session.get(
            '/__api_v2/datasets/',
            params=params,
            hooks=self._make_hook('No datasets in organisation.')
        ).json()

        return [
            dataset for dataset in dataset_response['data']
            # Filter on location type (take the first key only and check that this matches the location_type (??))
            if 'location' in dataset['attributes'] and next(
                iter(
                    dataset['attributes']['location'].keys()
                ),
                ''
            ).lower() == location_type
        ]

    def get_dataset(self, dataset_id: str) -> Dict:
        # todo -- add JWT arg (to be able to cache) and cache
        """
        Security: you cannot cache this at the moment because the JWT
        is not a parameter on this function.
        :param dataset_id:
        :return:
        """
        url = f'/__api_v2/datasets/{dataset_id}/'
        logger.debug(
            f"get_dataset - dataset_id '{dataset_id}'",
            extra={'url': url, 'dataset_id': dataset_id}
        )
        response = self.dlc_session.get(
            url,
            hooks=self._make_hook(f'Dataset {dataset_id} not found')
        )
        return response.json()['data']

    def get_datafile(self, datafile_id: str) -> dict:
        # todo -- add JWT arg (to be able to cache) and cache
        logger.debug(f"get_datafile - datafile_id '{datafile_id}'")
        return self.dlc_session.get(
            f'/__api/datafiles/{datafile_id}/',
            hooks=self._make_hook('Datafile {datafile_id} not found')
        ).json()


#
# Copyright (C) 2019 IHS Markit.
# All Rights Reserved
#
import dataclasses
import os
import logging
import hashlib
from typing import Optional
from botocore.exceptions import ClientError
from flask import current_app
from functools import wraps
from pymemcache.client.hash import HashClient
from injector import inject, provider, singleton as singleton_scope

from ..providers.base_provider import BaseModule
from ..clients.refreshing_hashclient import RefreshingHashClient


logger = logging.getLogger(__name__)


@dataclasses.dataclass
class Config:
    elasticache_config_node_url: str
    # Discovery timeout - probably will never need to be changed.
    time_to_timeout: int = 5
    # Defaults taken from the requests library were 10, 10. Under performance
    default_ttl: int = 3600


class MemcachedDependencyProvider(BaseModule):
    config_class = Config

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @singleton_scope
    @provider
    def memcached_client(self) -> HashClient:
        # NOTA BENE: DO NOT CHANGE THIS RETURN TYPE, even though we are not a HashClient, you will break injectors
        # as everywhere will be looking for the Provider of "HashClient" when we are really extending the type
        # by way of a HAS_A pass-through type: i.e. for our intents and purposes it acts like a HashClient.
        hashclient = RefreshingHashClient(self.config)
        return hashclient


def get_key_val(key, cache_func_name, *args, **kwargs):
    if key:
        # If a key parameter was provided, we do not do any skipping
        # of arguments. The caller has provided a key function so we
        # pass on all of the arguments and let the caller decide what
        # to do. This is important as the wrapper could be around a
        # static function that does not have `self` or `client` as the
        # first parameter, so skipping may cause us to miss an
        # important parameter.
        key_val = f'consumption-{cache_func_name}-' + \
                  key(*args, **kwargs)
    else:
        # Else, no key parameter was provided. This is treated as the
        # case for an annotation around a non-static or stateful
        # method. Such a method will start with `self` or `client`,
        # which is a value that breaks the caching so must be skipped.
        # The [1:] skips the `self` argument.
        a = '-'.join(str(a) for a in args[1:])
        k = '-'.join(f'{k}-{v}' for k, v in kwargs.items())
        # Use the commit ID so that when we do a deployment we start with a fresh cache.
        # This is useful for development.
        commit_id = os.environ.get(
            'CI_COMMIT_SHORT_SHA',
            'unknown'
        )
        key_val = f'consumption-{cache_func_name}-args:{a}:kwargs-{k}-' \
                  f'commit_id:{commit_id}'

    return key_val


def memcached_method(
    cache_getter,
    cache_func_name: Optional[str] = None,
    key=None,
    ttl=None,
    force_cache_refresh: bool = False,
):
    """
    Cache a call.

    :param cache_getter:
    :param key: func for generating the key used on the cache.

        If a key parameter was provided, we do not do any skipping
        of arguments. The caller hasprovided a key function so we
        pass on all of the arguments and let the caller decide what
        to do. This is important as the wrapper could be around a
        static function that does not have `self` or `client` as the
        first parameter, so skipping may cause us to miss an
        important parameter.

        Else, no key parameter was provided. This is treated as the
        case for an annotation around a non-static or stateful
        method. Such a method will start with `self` or `client`,
        which is a value that breaks the caching so must be skipped.
        The [1:] skips the `self` argument.

    :param ttl: Override the `ttl` found in the config file.
    :param cache_func_name: a name to add to the cache key so that:
     1. two functions with similar arguments do not have a name clash.
     2. if a property suce as the common prefixes have changes, then
     including them in the cache key will cause us to not hit the old cache
     key during pagination.
    :param force_cache_refresh: When False, will use standard cache
            behaviour of checking if a value is in cache, then return from
            cache, else  make a real call. When True, will make a real call
            and replace the values in the cache.
    :return: List of S3 files.
    """
    def decorator(func):

        @wraps(func)
        def wrapper(*args, **kwargs):
            nonlocal ttl
            nonlocal cache_func_name
            if args:
                # a class needing to be initialized or a function needing to be called to get the client
                mem_cache: HashClient = cache_getter(args[0])
            else:
                # a hashclient or equivalent with get/set methods
                mem_cache: HashClient = cache_getter()

            if cache_func_name is None:
                # If you are using cachedmethod as a decorator, then we can
                # get the name of the function the decorator is wrapping.
                # Otherwise, you need to explicitly provide a name.
                cache_func_name = func.__name__

            key_val = get_key_val(key, cache_func_name, *args, **kwargs)
            # TODO scott: in python 3.9 they add the parameter `usedforsecurity`. When we upgrade
            #  to Python 3.9 then set the parameter to False for a speed boost as we only use
            #  this for generating cache keys.
            key_hashed = hashlib.sha1(key_val.encode()).hexdigest()
            if len(key_hashed) > 250:
                # Guard against sending a key that is too long for memcached (limit 250
                # characters).
                #   SHA1 produces 40 characters hex encoded.
                #   MD5 produces 32 characters hex encoded.
                #   SHA-256 produces 64 characters hex encoded.
                logger.error(
                    'key_hashed is too long to be used as a key in memcached'
                )

            def _real_call(mem_cache_expiry, reason):
                logger.debug(f'Make real call to {func} - {reason}')
                # Make a real call.
                result_from_real_call = func(*args, **kwargs)

                # Save/replace the result in the cache.
                # Do not use memcached `.add` as it fails if the key already
                # exists on server. Use `set` instead.
                result_code = None
                if mem_cache:
                    try:
                        if not mem_cache_expiry:
                            mem_cache_expiry = current_app.dependencies.memcached.config.default_ttl

                        result_code = mem_cache.set(
                            key_hashed,
                            result_from_real_call,
                            expire=mem_cache_expiry
                        )
                    except Exception as e:
                        logging.error(e)

                if mem_cache and not result_code:
                    logger.warning(
                        'Failed to set cache value!',
                        extra={
                            'key_hashed': key_hashed,
                            'ttl': mem_cache_expiry,
                        }
                    )

                return result_from_real_call

            force_cache_refresh_override = os.environ.get('CACHE_DISABLED', force_cache_refresh)
            if force_cache_refresh_override:
                return _real_call(ttl, "because the ttl_cache is being forced to refresh")
            else:
                if mem_cache:
                    result_from_cache = mem_cache.get(key_hashed)
                    if result_from_cache:
                        # Result already in cache, so return it.
                        logger.debug(
                            f'Cache hit {key_val} - hash {key_hashed[:250]}'
                        )

                        if type(result_from_cache) != dict or not result_from_cache.get("error"):
                            return result_from_cache
                        else:
                            if result_from_cache.get("error") == "CLIENT_ERROR":
                                raise ClientError(result_from_cache.get("response"), result_from_cache.get("operation_name"))
                            else:
                                raise Exception()
                    else:
                        # Result not in cache, make a real call.
                        logger.debug(
                            f'Cache miss {key_val} - hash {key_hashed[:250]}'
                        )
                        return _real_call(ttl, "because the value is not in the ttl_cache")
                else:
                    # there is no cache
                    return _real_call(ttl, "because there is no ttl_cache attached")

        return wrapper

    return decorator





import os
from typing import Optional
from urllib.parse import urljoin
import requests
from flask import Blueprint, request, Flask
from injector import Injector

import pytest
from unittest.mock import MagicMock

from ..authentication.middleware import ServiceMiddleware


def mock_middleware_server(views: Blueprint):

    def create_app():
        os.environ["CATALOGUE_API_URL"] = "test"
        application = Flask(__name__)
        application.wsgi_app = ServiceMiddleware(application.wsgi_app)
        application.injector = Injector()

        # we need to access the injector in the wsgi (early), so we ascribe it to the object too
        application.register_blueprint(views)
        application.wsgi_app.injector = application.injector

        return application

    return create_app()


@pytest.fixture
def disconnected_client():

    views = Blueprint('middleware', __name__)
    HTTP_METHODS = ['GET', 'HEAD', 'POST', 'PUT', 'DELETE', 'CONNECT', 'OPTIONS', 'TRACE', 'PATCH']

    @views.route(
        '/',
        methods=HTTP_METHODS
    )
    @views.route(
        '/<path:optional_path>',
        methods=HTTP_METHODS
    )
    def test_route(optional_path: Optional[str] = None):
        resp = MagicMock()
        resp.status_code = 200
        resp.headers = {}
        resp.content = "HELLO, WORLD"
        return resp.content, resp.status_code, resp.headers.items()

    @views.route(
        '/direct_test',
        methods=HTTP_METHODS
    )
    def test_direct_route(optional_path: Optional[str] = None):
        resp = MagicMock()
        resp.status_code = 403
        resp.headers = {}
        resp.content = "BE GONE, DRAGON"
        return resp.content, resp.status_code, resp.headers.items()

    app = mock_middleware_server(views)
    with app.test_client() as client:
        yield client


#
# Copyright (C) 2019 IHS Markit.
# All Rights Reserved
#
import logging
try:
    # these are only present for services, not for the SDK, and we do not wish to impose these in requirements.txt
    from injector import inject
except ModuleNotFoundError:
    # set up an Identity decorator for @inject
    def inject(function): # type: ignore
        def wrapper(*args, **kwargs):
            return function(*args, *kwargs)
        return wrapper
from typing import Iterator, Dict, Any, Optional, Iterable

from ..consumption_service import FileService
from ..dlc_service import DlcService
from ..arrow_service.handlers.parquet import ParquetHandler
from ..arrow_service.handlers.csv import CsvHandler
from ..arrow_service.exceptions import UnsupportedDatasetFormat


class ArrowService:

    logger = logging.getLogger(__name__)

    @inject
    def __init__(
        self,
        file_service: FileService,
        dlc_service: DlcService,
        **kwargs,
    ):
        self.file_service = file_service
        self.dlc_service = dlc_service

        self.format_handlers: Dict[str, Any] = {
            'parquet': ParquetHandler,
            'csv': CsvHandler
        }

        self.logger.debug(
            'arrow service init',
        )

    # XXX TODO move into dataset class
    def _get_dataset_handler(self, dataset):
        # TODO cache get dataset calls. Each one takes roughly 250ms.
        # Perhaps provide some sort of OO interface like Dataset.from_id()
        # that would keep a class registry of datasets currently loaded.

        data_format = dataset['attributes']['data_format'].lower()
        dataset_name = dataset['attributes']['name']
        dataset_id = dataset['id']

        self.logger.debug(
            'Dataset in handler',
            extra={
                'dataset_name': dataset_name,
                'dataset_id': dataset_id,
                'data_format': data_format,
            }
        )

        try:
            handler_cls = self.format_handlers[data_format]
        except KeyError as e:
            supported_formats = ','.join(self.format_handlers.keys()).lower()
            raise UnsupportedDatasetFormat(
                details=(
                    f'Only {supported_formats} currently '
                    f'supports dataframe streaming via consumption. The '
                    f'dataset {dataset_id} - {dataset_name} is of type: '
                    f'{data_format} and can not be streamed.'
                )
            ) from e

        return handler_cls

    def dataset_to_ipc_arrow_stream(
        self,
        dataset: Dict,
        nrows=None,
        partitions: Optional[Iterable] = None,
        order_by_latest: Optional[bool] = None,
    ) -> Iterator[bytes]:

        self.logger.debug(
            'dataset_to_ipc_arrow_stream',
        )
        # If order_by_latest is True or False then the files should be
        # returned ordered.

        files = self.file_service.get_files_from_dataset(
            dataset, partitions=partitions, order_by_latest=order_by_latest,
            # Hidden files can be needless metadata
            skip_hidden_files=True,
            skip_special_hadoop_files=True,
        )

        handler = self._get_dataset_handler(dataset)
        yield from handler(files=files, order_by_latest=order_by_latest).\
            handle(nrows=nrows)



#
# Copyright (C) 2019 IHS Markit.
# All Rights Reserved
#
import itertools
import os
import logging
import dataclasses
import sys
import numpy
import pyarrow as pa
from typing import Optional, Iterator, List
from pyarrow import (
    RecordBatchStreamWriter, RecordBatch
)
from pyarrow.lib import ArrowException
from pyarrow.parquet import ParquetFile

from ...arrow_service.exceptions import (NoFilesInDataset, NonParquetFilesInDataset)
from ...arrow_service.handlers.dequeue_file import DequeueFile
from ...arrow_service.handlers.stream_errors import handle_arrow_stream_errors
from ....services.exceptions import ServiceException, SchemaMismatchException
from ....adapters.s3_io_adapters import ServiceIO


@dataclasses.dataclass
class Config:
    dd_profiler_enabled: bool = \
        True if os.environ.get('DDTRACE_PROFILER_ENABLED', False) == 'True' \
        else False
    # Overridden in tests to False for ease of testing.
    # In the prod code this is True so that we can free
    # memory as soon as possible.
    close_s3_files: bool = True


class ParquetHandler:

    logger = logging.getLogger(__name__)

    def __init__(self, files: Iterator[ServiceIO],
                 order_by_latest: Optional[bool] = True,
                 **kwargs):
        self.config = Config(**kwargs)

        self.logger.debug(
            'ParquetHandler init',
            extra={
                'config': dataclasses.asdict(self.config),
                'order_by_latest': order_by_latest,
            }
        )
        initial_file = None
        self.initial_file_path = 'not set'

        try:

            if order_by_latest is True:
                # The parquet schema handling depends on us knowing which
                # is the latest .parquet file. The list of files has to
                # be ordered in some way (ascending or descending),
                # it cannot be un-ordered.
                self.logger.debug(
                    'Use the schema from the first file in the list.'
                )
            else:
                # Use the schema from the first file in the list. This
                # will trigger a listing of the files so will be slower
                # to get the first bytes of data back to the client.

                # For the case of order_by_latest None, the default of S3
                # seems to be to return sorted ascending alphabetically,
                # ASSUMPTION this may be close enough. If not, we will have
                # to change the default in dataframe/views.py from None to
                # either True or False.
                self.logger.debug(
                    'Use the schema from the last file in the list.'
                )

                # TODO for the additive schema we need to know the final
                #  file and will use that as the schema for the whole
                #  response. We need to cast to type List to action the
                #  iterator so we can flip the list. Alternative to
                #  caching the whole list is to instead cache only the
                #  S3File of the last file and process that file
                #  immediately (so nginx is able to return quickly),
                #  reading the rest of the files straight from the S3
                #  list not from cache and therefore we can keep `files`
                #  as type iterator. This would also require a crawler to
                #  populate cache with the latest as sorting takes time.
                files = iter(reversed(list(files)))

            # This will raise StopIteration if there are no files in the dataset.
            initial_file = next(files)

            # First time only we get the schema. Unfortunately we need to do
            # this upfront as we need to know the schema before going into the
            # RecordBatchStreamWriter loop.
            self.initial_file_path = initial_file.path
            self.logger.debug(
                'initial_file_path',
                extra={
                    'path': self.initial_file_path,
                }
            )
            self.__validate_file_extension(initial_file)

            initial_parquet_file: ParquetFile = initial_file. \
                to_parquet_file()

        except StopIteration as e:
            message = 'No data in dataset. Unable to stream.'
            self.logger.info(message)
            raise NoFilesInDataset(message) from e

        except (OSError, UnicodeDecodeError, ArrowException) as e:
            message = 'Error for initial file. ' \
                      'Unable to stream.'
            self.logger.warning(
                message,
                exc_info=e,
                extra={'path': self.initial_file_path}
            )
            raise NonParquetFilesInDataset(
                'Non parquet file found in dataset. Path: '
                f'{self.initial_file_path}'
            ) from e

        except NonParquetFilesInDataset:
            raise

        except Exception as e:
            message = 'The file is corrupt, ' \
                      'not a .parquet like file or cannot be read. ' \
                      'Please report the problem to the datalake ' \
                      'platform team at DataLake-Support_L1@ihsmarkit.com. ' \
                      'Please include the full error message including ' \
                      'the request_id.'
            self.logger.exception(message)
            raise ServiceException(message) from e

        # Publicly accessed

        self.schema: pa.Schema = initial_parquet_file.schema.to_arrow_schema()

        self.logger.debug(
            'Parquet columns mapped to pyarrow schema names',
            extra={
                'schema_names': self.schema.names
            }
        )
        self.logger.info(
            'Parquet columns mapped to pyarrow schema types',
            extra={
                'schema_names': self.schema.types
            }
        )

        if order_by_latest is True:
            self._files = itertools.chain([initial_file], files)
        else:
            # The list had been flipped from ascending to descending,
            # so now flip it back. This is clunky and would be better if
            # we only cached the initial schema file.
            self._files = itertools.chain(
                reversed(list(files)),
                [initial_file]
            )

    def __validate_file_extension(self, s3_file: ServiceIO):
        if not s3_file.path.endswith('.parquet'):
            self.logger.warning(
                'Non parquet file found in dataset.',
                extra={
                    'path': s3_file.path,
                }
            )
            raise NonParquetFilesInDataset(
                f'Non parquet file found in dataset. Path: {s3_file.path}'
            )

    @staticmethod
    def _validate_file_size(s3_file: ServiceIO):
        """File size must be greater than zero."""
        if hasattr(s3_file, 'size'):
            return s3_file.size > 0
        else:
            return False

    def _get_record_batches_from_files(self) -> Iterator[RecordBatch]:
        for file_number, s3_file in enumerate(self._files):
            extra_log = {'file_number': file_number, 'path': s3_file.path}

            if not self._validate_file_size(s3_file):
                # Skip zero size files.
                self.logger.warning(
                    'File has been skipped, Parquet file size is 0 bytes.',
                    extra=extra_log
                )
                if self.config.close_s3_files:
                    # Do not hold onto the file contents, close to
                    # free the memory.
                    s3_file.close()
                continue

            self.__validate_file_extension(s3_file)

            parquet_file: ParquetFile = s3_file.to_parquet_file()

            extra_log['num_row_groups'] = parquet_file.num_row_groups

            self.logger.debug(
                f"Reading from parquet, file number '{file_number}'",
                extra=extra_log
            )

            if self.config.dd_profiler_enabled:
                # Log to see the memory usage.
                self.logger.debug(
                    f's3_file size: {sys.getsizeof(s3_file)} bytes '
                    'before closing',
                    extra=extra_log
                )
                self.logger.debug(
                    f'parquet_file size: {sys.getsizeof(parquet_file)}'
                    ' bytes ',
                    extra=extra_log
                )

            current_arrow_schema = parquet_file.schema_arrow.\
                remove_metadata()

            # If the current schema is not a subset
            # TODO scott: we should also be able to work with transposed
            # columns. This will also be necessary for when it is additive
            # but they have added a column in the middle instead of at the end!
            is_schema_ok = (
                list(current_arrow_schema) ==
                list(self.schema)[:len(current_arrow_schema)]
            )
            if not is_schema_ok:
                raise SchemaMismatchException(
                    'Tried to process a file with a schema that does not '
                    'match the schema of the latest. We '
                    'require that all .parquet files have an additive '
                    'schema.'
                    "\n\nLatest file's schema:"
                    f"\n\n{repr(self.schema.remove_metadata())}"
                    "\n\nLatest file's path:"
                    f'\n\n{self.initial_file_path}'
                    f"\n\nCurrent file's schema:"
                    f"\n\n{repr(current_arrow_schema)}"
                    "\n\nCurrent file's path:"
                    f'\n\n{s3_file.path}'
                )
            self.logger.debug('schema OK', extra=extra_log)

            unused_fields: List[pa.Field] = \
                list(self.schema)[len(current_arrow_schema):]
            extra_log['len unused_fields'] = len(unused_fields)

            for row_group_number in range(parquet_file.num_row_groups):

                def get_batches_with_additive_columns() -> List[RecordBatch]:
                    extra_log['row_group_number'] = row_group_number
                    self.logger.debug('Before read_row_group', extra=extra_log)

                    # Content of the row group as a table of columns.
                    row_group: pa.Table = parquet_file. \
                        read_row_group(i=row_group_number)

                    self.logger.debug(
                        'Able to read_row_group from parquet_file',
                        extra=extra_log
                    )

                    if self.config.dd_profiler_enabled:
                        # Log to see the memory usage of each batch.
                        self.logger.debug(
                            f'row_group size: {sys.getsizeof(row_group)} bytes '
                            'before additive schema',
                            extra=extra_log
                        )

                    for field in unused_fields:
                        # TODO scott: This may need a change to set_column if
                        #   we have transposed columns, as we would need to
                        #   append in a specific order.
                        row_group = row_group.append_column(
                            field_=field,
                            # Append and empty array of the correct type and
                            # size to fill in the missing column.
                            column=pa.array(
                                numpy.empty(row_group.num_rows, dtype=object),
                                field.type
                            )
                        )

                    if self.config.dd_profiler_enabled:
                        # Log to see the memory usage of each batch.
                        self.logger.debug(
                            f'row_group size: {sys.getsizeof(row_group)} bytes '
                            'after additive schema',
                            extra=extra_log
                        )

                    self.logger.debug(
                        'Convert row_group to_batches after additive schema',
                        extra=extra_log
                    )

                    return row_group.to_batches()

                # row group no longer referenced after the call so
                # the memory should automatically be freed.
                batches: List[RecordBatch] = \
                    get_batches_with_additive_columns()

                self.logger.debug(
                    'After additive schema to_batches',
                    extra=extra_log
                )

                # At this point we ask for the first batch from the
                # batches iterator, our memory usage to increase massively
                # and does not reduce, even when we leave the outer loop
                # and if we explicitly call gc.collect(1).
                for batch in batches:
                    yield batch
                    if self.config.dd_profiler_enabled:
                        # Log to see the memory usage of each batch.
                        self.logger.debug(
                            f'Batch size: {sys.getsizeof(batch)} bytes',
                            extra=extra_log
                        )

            if self.config.close_s3_files:
                if self.config.dd_profiler_enabled:
                    # Log to see the memory usage of each batch.
                    self.logger.debug(
                        f's3_file size: {sys.getsizeof(s3_file)} bytes before '
                        'closing',
                        extra=extra_log
                    )

                # Do not hold onto the file contents.
                s3_file.close()

                # Log to see the memory usage of each batch.
                self.logger.debug(
                    f'Closed s3_file. Do not hold onto the file contents.',
                    extra=extra_log
                )

    def handle(self, nrows: Optional[int] = None) -> Iterator[bytes]:

        record_batches: Iterator[RecordBatch] = \
            self._get_record_batches_from_files()
        sink = DequeueFile()

        try:
            # double dispatch so that we don't yield in finally
            yield from self._inner_handle(sink=sink, record_batches=record_batches, nrows=nrows)
        except Exception:
            yield from sink.dequeue()
            raise

        yield from sink.dequeue()

    def _inner_handle(
        self,
        sink,
        record_batches: Iterator[RecordBatch],
        nrows: Optional[int] = None,
    ) -> Iterator[bytes]:

        rows_iterated = 0

        self.logger.debug(
            'Handling parquet file',
            extra={'nrows': nrows}
        )

        with handle_arrow_stream_errors(
            RecordBatchStreamWriter(sink, self.schema)
        ) as writer:
            for batch in record_batches:
                if nrows and (rows_iterated + batch.num_rows) > nrows:
                    self.logger.debug(
                        'Truncating last batch',
                        extra={
                            'nrows': nrows,
                            'rows_iterated': rows_iterated,
                            'batch num_rows': batch.num_rows,
                        }
                    )
                    batch_truncated = batch.slice(0, (nrows - rows_iterated))
                    writer.write_batch(batch_truncated)
                    yield from sink.dequeue()
                    rows_iterated = rows_iterated + batch_truncated.num_rows
                    break
                else:
                    writer.write_batch(batch)
                    yield from sink.dequeue()
                    rows_iterated = rows_iterated + batch.num_rows

        self.logger.debug(
            'Completed iteration',
            extra={'nrows': nrows, 'rows_iterated': rows_iterated}
        )

    def __del__(self):
        """Destructor."""
        self.logger.debug('Destructor called on parquet handler')



import dataclasses
import logging
from injector import singleton as singleton_scope, provider
from requests import Session
from urllib3 import Retry

from ..providers.base_provider import BaseModule
from ..adapters.session_adapters import DlcAdapter


@dataclasses.dataclass
class Config:
    api_root: str
    # Defaults taken from the requests library were 10, 10. Under performance
    # tests with 250 concurrent users and 30 pods, 10 was not enough.
    pool_connections: int = 20
    pool_maxsize: int = 20
    max_retries: int = 10


class JWTIdentityDlcClient(Session):

    logger = logging.getLogger(__name__)

    def __init__(self, config):
        super().__init__()
        self.config = config

        adapter = DlcAdapter(
            pool_connections=self.config.pool_connections,
            pool_maxsize=self.config.pool_maxsize,
            # https://urllib3.readthedocs.io/en/latest/reference/
            # urllib3.util.html#urllib3.util.retry.Retry
            max_retries=Retry(
                total=self.config.max_retries,
                status=self.config.max_retries,
                status_forcelist=[500, 502, 504],
                method_whitelist=Retry.DEFAULT_METHOD_WHITELIST,
                backoff_factor=1,
            ),
            pool_block=True,
        )

        self.logger.debug(
            'JWTIdentityDlcClient session'
            f" mount prefix '{self.config.api_root}'"
        )
        self.mount(self.config.api_root, adapter)

    def request(self, method, url, *args, **kwargs):
        # url = urljoin(self.config.api_root, url) # this isnt going to work for a local case
        url = "".join([self.config.api_root, url])
        self.logger.debug(
            'JWTIdentityDlcClient request using connection pool',
            extra={'url': url}
        )
        return super().request(method, url, *args, **kwargs)


class JWTIdentityDependencyProvider(BaseModule):
    config_class = Config

    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        self.session = JWTIdentityDlcClient(config=self.config)

    @singleton_scope
    @provider
    def auth_client(self) -> JWTIdentityDlcClient:
        return self.session


from injector import inject
import logging

from ..providers.identity_provider import JWTIdentityDlcClient
from ..services.abstract_external_service import ExternalService

logger = logging.getLogger(__name__)


class IdentityService(ExternalService):
    """
    Data Lake interface service, for interfacing
    with said interface.

    The JWTIdentityDlcClient should ONLY be responsible
    for authentication and connection pooling. This class
    should be responsible for implementation of high level
    interfacing with the interface component.
    """
    @inject
    def __init__(
        self,
        identity_session: JWTIdentityDlcClient
    ):
        logger.debug('Init IdentityService')
        self.identity_session = identity_session

    def get_visible_organisations(self):
        return self.identity_session.get(
            '/__api_v2/organisations/visible',
            hooks=self._make_hook('No visible organisations.')
        ).json()

    def me(self):
        return self.identity_session.get(
            '/__api_v2/me',
            hooks=self._make_hook('Cannot access identity/me.')
        ).json()


import os
import logging
from cachetools import TTLCache
from injector import singleton as singleton_scope, provider

from ..providers.base_provider import BaseModule

logger = logging.getLogger(__name__)


class InMemoryTTLCacheDependencyProvider(BaseModule):

    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        maxsize: int = int(os.environ.get(
            'ASSUME_ROLE_CACHE_MAX_SIZE',
            '16'
        ))
        ttl: int = int(os.environ.get(
            'ASSUME_ROLE_CACHE_TTL_SECONDS',
            '900'  # 15 minutes
        ))

        if not os.environ.get('DISABLE_CACHE', None):
            logger.info("Created TTL Cache")
            self.session = TTLCache(maxsize=maxsize, ttl=ttl)
        else:
            logger.warning("TTL Cache is disabled")
            self.session = None

    @singleton_scope
    @provider
    def ttl_cache(self) -> TTLCache:
        return self.session


'''
 Do not remove this alias file

 Its only purpose is to provide backwards compatibility
 for users of the Builder classes, which have been moved to components.datasets

'''
import warnings

warnings.warn("DatasetBuilder and DatasetLocationBuilder have now moved, you can still use them here, but "
              "please update your imports as below, as they will deprecated in the future (1.20). Change from:\n\n"
              "from dli.client.builders import DatasetLocationBuilder \n"
              "to \n"
              "from dli.components.dataset import DatasetLocationBuilder"
              "\n"
              "\n"
              "from dli.client.builders import DatasetBuilder \n"
              "to \n"
              "from dli.components.dataset import DatasetBuilder\n\n"
              )
from dli.components.dataset import DatasetLocationBuilder
from dli.components.dataset import DatasetBuilder


import io
import logging
from abc import ABCMeta
from pyarrow.parquet import ParquetFile

logger = logging.getLogger(__name__)


class ServiceIO(io.RawIOBase, metaclass=ABCMeta):
    backend_type: str
    path: str
    metadata: dict
    size: int

    def to_parquet_file(self) -> ParquetFile:
        # Note the source works when set to the ServiceIO instance, do not
        # just pass in the path.
        return ParquetFile(
            source=self,
            # Buffer size > 0 tells pyarrow to use buffering when
            # deserializing individual column chunks. Setting this to a low
            # value reduces the memory usage and amount of data read. Not
            # setting this will cause the whole column chunk to be read into
            # memory unbuffered.
            buffer_size=0,
            # TODO scott: memorymap = True did not help memory usage with
            #   Rollo's pyarrow 0.16.1, but we haven't tried again in
            #   pyarrow 2.0.0.
            # Tried with buffer_size >0 in pyarrow 2.0.0 but got the error
            # message:
            # File "pyarrow/error.pxi", line 99, in pyarrow.lib.check_status
            # OSError: IOError: An HTTP Client raised an unhandled exception:
            # cannot switch to a different thread. Detail: Python exception:
            # HTTPClientError
        )


class S3File(ServiceIO):
    """
    Abstracts over an S3 object summary to return a *seekable*
    io object. Using the S3 Range header to emulate the seek
    method.
    """

    # NOTE These are additional methods on the
    # file object used later on in serializing metadata
    # all backends (for now just S3 - but  in future
    # and datastore like google cloud azure, SAN filesystems)
    # are expected to implement these methods.
    backend_type = 's3'

    def __init__(
        self,
        key_dict: dict = None,
        bucket_name: str = None,
        get_object=None,
    ):
        self.key_dict = key_dict
        # There is a non trivial overhead to using the OO resource.
        # probably because boto service definitions are defined in
        # json, there's an IO overhead in calling functions. Fetching
        # and binding this ahead of time improves things.
        self.get_object_func = get_object
        self.bucket_name = bucket_name
        self.position = 0

    @property
    def key(self):
        return self.key_dict['Key']

    @property
    def metadata(self):
        return {
            'e_tag': self.key_dict['ETag'].strip('"'),  # e_tag is md5 of file
            'size': self.key_dict['Size'],
            'last_modified': self.key_dict['LastModified']
        }

    @property
    def size(self):
        return int(self.key_dict['Size'])

    @property
    def path(self):
        return self.key_dict['Key']

    def __repr__(self):
        return '<{} bucket_name={} path={}>'.format(
            type(self).__name__,
            self.bucket_name,
            self.path
        )

    def tell(self):
        return self.position

    def seek(self, offset, whence=io.SEEK_SET):
        if whence == io.SEEK_SET:
            self.position = offset
        elif whence == io.SEEK_CUR:
            self.position += offset
        elif whence == io.SEEK_END:
            self.position = self.size + offset
        else:
            raise ValueError(
                'invalid whence ({}, should be {}, {}, {})'.format(
                    whence, io.SEEK_SET, io.SEEK_CUR, io.SEEK_END
                )
            )

        return self.position

    def seekable(self):
        return True

    def read(self, size=-1):
        if size == -1:
            # Read to the end of the file
            logger.debug('Read to the end of the S3File')
            range_header = 'bytes={}-'.format(self.position)
            self.seek(offset=0, whence=io.SEEK_END)

        else:
            new_position = self.position + size

            # AWS doesn't support starting bytes positions
            # at the end of the file (which would return an
            # empty bytes string anyway). i.e this
            # self.get_func(Range='bytes={self.size}-')
            # will fail
            if self.position >= self.size:
                return b''

            # If we're going to read beyond the end of the object, return
            # the entire object.
            if new_position >= self.size:
                return self.read()

            range_header = 'bytes={}-{}'.format(
                self.position, new_position - 1
            )

            self.seek(offset=size, whence=io.SEEK_CUR)

        obj = self.get_object_func(
            Bucket=self.bucket_name,
            Key=self.key,
            Range=range_header,
        )

        return obj['Body'].read()

    def readable(self):
        return True

    def readall(self):
        return self.read()

    def readinto(self, b):
        """
        Read bytes into a pre-allocated, writable bytes-like object b

        :param b: Bytes buffer, might be a bytearray. Cannot use typing on
        this method as we get a mypy error message that contradicts the
        source code:

        Argument 1 of "readinto" is incompatible with supertype "RawIOBase";
        supertype defines the argument type as
        "Union[bytearray, memoryview, array[Any], mmap]".

        :return: Return the number of bytes read. If no bytes are available,
        None is returned.
        """
        # Use a memoryview because it slices without copying the data, which
        # makes it fast/memory efficient.
        # https://julien.danjou.info/high-performance-in-python-with-zero-copy
        # -and-the-buffer-protocol/
        view = memoryview(b)
        # Try to read enough of the file to fill up the buffer.
        data = self.read(len(view))
        # Handle the case of asking for more bytes than there are in the file.
        number_of_bytes_read = len(data)
        view[:number_of_bytes_read] = data
        return number_of_bytes_read

    def write(self, b):
        raise NotImplementedError(
            'We should not be writing, the object should be read-only'
        )


#
# Copyright (C) 2020 IHS Markit.
# All Rights Reserved
#
import re
from collections import namedtuple
from typing import Dict


class _AttributeAdapter(object):

    def get_field_names_to_attributes(self) -> Dict[str, str]:
        field_names_to_attributes = dict(
            [
                (field_name, getattr(self, field_name)) for field_name in dir(self)
                  if not field_name.startswith('__')
                  and not field_name.startswith('_')
                  and not callable(getattr(self, field_name))
            ]
        )

        if 'datasetId' not in field_names_to_attributes and \
                'dataset_id' in field_names_to_attributes:
            # DL-6445: For backwards compatibility re-add the old attribute name.
            field_names_to_attributes['datasetId'] = field_names_to_attributes['dataset_id']

        return field_names_to_attributes

    @staticmethod
    def __fun(clazz, field_name) -> str:
        return field_name + "='" + str(getattr(clazz, field_name)) + "'"

    def __str__(self):
        sanitised_fields = [self.__fun(self, field_name) for field_name in dir(self) if
                 not field_name.startswith('__') and not field_name.startswith(
                     '_') and not callable(getattr(self, field_name))]

        fields_str: str = ", ".join(sanitised_fields)
        return f"{self.__class__.__name__}({fields_str})"

    def as_dict(self):
        return self.get_field_names_to_attributes()

    def _asdict(self):
        # Support alternative spelling.
        return self.as_dict()


def to_camel_case(snake_str):
    split = snake_str.split('_')
    camel_case_parts = [split[0].lower()] + [s.title() for s in split[1:]]
    return ''.join(camel_case_parts)


def to_camel_cased_dict(dictionary):
    return {to_camel_case(key): value for (key, value) in dictionary.items()}


def to_snake_case(s):
    # need to sanitise the string as in some cases the key might look like
    # 'Data Access' with a space.
    _reg = re.compile(r'(?!^)(?<!_)([A-Z])')
    return _reg.sub(r'_\1', s.replace(' ', '_')).lower()


def object_to_namedtuple(obj):
    e = obj.as_dict()
    namedtuple_obj = namedtuple(obj.__class__.__name__, sorted(e.keys()))(**e)
    return namedtuple_obj


def to_snake_case_keys(dictionary: dict):
    dictionary_snake_case = {to_snake_case(key): value for (key, value) in dictionary.items()}
    return dictionary_snake_case


#
# Copyright (C) 2020 IHS Markit.
# All Rights Reserved
#
import os
from urllib.parse import urlparse, ParseResult


class _Environment:

    _catalogue_environment_name_map = {
        'catalogue.datalake.ihsmarkit.com': 'prod',
        'catalogue-testprod.datalake.ihsmarkit.com': 'testprod',
        'catalogue-uat.datalake.ihsmarkit.com': 'uat',
        'catalogue-uatus.datalake.ihsmarkit.com': 'uatus',
        'client-uat.datalake.ihsmarkit.com': 'uat2',
        'catalogue-dev.udpmarkit.net': 'dev',
        'catalogue-qa.udpmarkit.net': 'qa',
        'catalogue-qa2.udpmarkit.net': 'qa2',
        'catalogue-qa3.udpmarkit.net': 'qa3',
    }

    _catalogue_accounts_environment_map = {
        'catalogue.datalake.ihsmarkit.com': 'catalogue.datalake.ihsmarkit.com',
        'catalogue-testprod.datalake.ihsmarkit.com': 'catalogue-testprod.datalake.ihsmarkit.com',
        'catalogue-uat.datalake.ihsmarkit.com': 'catalogue-uat.datalake.ihsmarkit.com',
        'catalogue-uatus.datalake.ihsmarkit.com': 'catalogue-uatus.datalake.ihsmarkit.com',
        'client-uat.datalake.ihsmarkit.com': 'client-uat.datalake.ihsmarkit.com',
        'catalogue-dev.udpmarkit.net': 'catalogue-dev.udpmarkit.net',
        'catalogue-qa.udpmarkit.net': 'catalogue-qa.udpmarkit.net',
        'catalogue-qa2.udpmarkit.net': 'catalogue-qa2.udpmarkit.net',
        'catalogue-qa3.udpmarkit.net': 'catalogue-qa3.udpmarkit.net',
    }

    _catalogue_consumption_environment_map = {
        'catalogue.datalake.ihsmarkit.com': 'consumption.datalake.ihsmarkit.com',
        'catalogue-testprod.datalake.ihsmarkit.com': 'consumption-testprod.datalake.ihsmarkit.com',
        'catalogue-uat.datalake.ihsmarkit.com': 'consumption-uat.datalake.ihsmarkit.com',
        'catalogue-uatus.datalake.ihsmarkit.com': 'consumption-uatus.datalake.ihsmarkit.com',
        'client-uat.datalake.ihsmarkit.com': 'consumption-uat2.datalake.ihsmarkit.com',
        'catalogue-dev.udpmarkit.net': 'consumption-dev.udpmarkit.net',
        'catalogue-qa.udpmarkit.net': 'consumption-qa.udpmarkit.net',
        'catalogue-qa2.udpmarkit.net': 'consumption-qa2.udpmarkit.net',
        'catalogue-qa3.udpmarkit.net': 'consumption-qa3.udpmarkit.net',
    }

    _catalogue_sam_api_environment_map = {
        'catalogue.datalake.ihsmarkit.com': 'sam-api.ihsmarkit.com',
        'catalogue-testprod.datalake.ihsmarkit.com': 'sam-api.ihsmarkit.com',
        'catalogue-uat.datalake.ihsmarkit.com': 'sam-api.ihsmarkit.com',
        'catalogue-uatus.datalake.ihsmarkit.com': 'sam-api.ihsmarkit.com',
        'client-uat.datalake.ihsmarkit.com': 'sam-api.ihsmarkit.com',
        'catalogue-qa.udpmarkit.net': 'sam-api.samexternal.net',
        'catalogue-qa2.udpmarkit.net': 'sam-api.samexternal.net',
        'catalogue-qa3.udpmarkit.net': 'sam-api.samexternal.net',
        'catalogue-dev.udpmarkit.net': 'sam-api.samexternal.net',
    }

    _catalogue_sam_environment_map = {
        'catalogue.datalake.ihsmarkit.com': 'sam.ihsmarkit.com',
        'catalogue-testprod.datalake.ihsmarkit.com': 'sam.ihsmarkit.com',
        'catalogue-uat.datalake.ihsmarkit.com': 'sam.ihsmarkit.com',
        'catalogue-uatus.datalake.ihsmarkit.com': 'sam.ihsmarkit.com',
        'client-uat.datalake.ihsmarkit.com': 'sam.ihsmarkit.com',
        'catalogue-qa.udpmarkit.net': 'sam.samexternal.net',
        'catalogue-qa2.udpmarkit.net': 'sam.samexternal.net',
        'catalogue-qa3.udpmarkit.net': 'sam.samexternal.net',
        'catalogue-dev.udpmarkit.net': 'sam.samexternal.net',
    }

    _s3_proxy_environment_map = {
        'catalogue.datalake.ihsmarkit.com': 's3.datalake.ihsmarkit.com',
        'catalogue-testprod.datalake.ihsmarkit.com': 's3-testprod.datalake.ihsmarkit.com',
        'catalogue-uat.datalake.ihsmarkit.com': 's3-uat.datalake.ihsmarkit.com',
        'catalogue-uatus.datalake.ihsmarkit.com': 's3-uatus.datalake.ihsmarkit.com',
        'client-uat.datalake.ihsmarkit.com': 's3-uat2.datalake.ihsmarkit.com',
        'catalogue-qa.udpmarkit.net': 's3-qa.udpmarkit.net',
        'catalogue-qa2.udpmarkit.net': 's3-qa2.udpmarkit.net',
        'catalogue-qa3.udpmarkit.net': 's3-qa3.udpmarkit.net',
        'catalogue-dev.udpmarkit.net': 's3-dev.udpmarkit.net',
    }

    _presto_environment_map = {
        'catalogue.datalake.ihsmarkit.com': 'sql.datalake.ihsmarkit.com',
        'catalogue-uat.datalake.ihsmarkit.com': 'presto-uat.datalake.ihsmarkit.com',
        'client-uat.datalake.ihsmarkit.com': 'sql-uat2.datalake.ihsmarkit.com',
        'catalogue-qa.udpmarkit.net': 'presto-qa.udpmarkit.net',
        'catalogue-qa2.udpmarkit.net': 'presto-qa2.udpmarkit.net',
        'catalogue-dev.udpmarkit.net': 'presto-dev.udpmarkit.net',
    }

    def __init__(self, api_root):
        """
        Class to manage the different endpoints

        :param str root_url: The root url of the catalogue
        """

        __catalogue_parse_result = urlparse(api_root)

        self.environment_name = self._catalogue_environment_name_map.get(
            __catalogue_parse_result.netloc, 'unknown'
        )

        # allow URLs to be overridden by ENV vars (SAM is fixed as Prod or Non-Prod, specific to env)
        # make sure to have the api_root of the SDK still set to the respective environment however

        if os.environ.get("DATA_LAKE_INTERFACE_URL"):
            self.catalogue = os.environ.get("DATA_LAKE_INTERFACE_URL")
        else:
            self.catalogue = ParseResult(
                __catalogue_parse_result.scheme, __catalogue_parse_result.netloc, '', '', '', ''
            ).geturl()

        if os.environ.get("DATA_LAKE_ACCOUNTS_URL"):
            self.accounts = os.environ.get("DATA_LAKE_ACCOUNTS_URL")
        else:
            __accounts_host = self._catalogue_accounts_environment_map.get(
                __catalogue_parse_result.netloc
            )

            self.accounts = ParseResult(
                __catalogue_parse_result.scheme, __accounts_host, '', '', '', ''
            ).geturl()

        if os.environ.get("CONSUMPTION_API_URL"):
            self.consumption = os.environ.get("CONSUMPTION_API_URL")
        else:
            __consumption_host = self._catalogue_consumption_environment_map.get(
                __catalogue_parse_result.netloc
            )

            self.consumption = ParseResult(
                __catalogue_parse_result.scheme, __consumption_host, '', '', '', ''
            ).geturl()

        if os.environ.get("S3_API_URL"):
            self.s3_proxy = os.environ.get("S3_API_URL")
        else:
            __s3_proxy_host = self._s3_proxy_environment_map.get(
                __catalogue_parse_result.netloc
            )

            self.s3_proxy = ParseResult(
                __catalogue_parse_result.scheme, __s3_proxy_host, '', '', '', ''
            ).geturl()

        if os.environ.get("SAM_URL"):
            self.sam = os.environ.get("SAM_URL")
        else:
            __sam_host = self._catalogue_sam_environment_map.get(
                __catalogue_parse_result.netloc
            )

            self.sam = ParseResult(
                'https', __sam_host, '', '', '', ''
            ).geturl()

        if os.environ.get("SAM_API_URL"):
            self.sam_api = os.environ.get("SAM_API_URL")
        else:
            __sam_api_host = self._catalogue_sam_api_environment_map.get(
                __catalogue_parse_result.netloc
            )

            self.sam_api = ParseResult(
                'https', __sam_api_host, '', '', '', ''
            ).geturl()

        if os.environ.get("SQL_URL"):
            self.sql = os.environ.get("SQL_URL")
        else:
            __presto_host = self._presto_environment_map.get(
                __catalogue_parse_result.netloc
            )

            self.sql = ParseResult(
                'https', __presto_host, '', '', '', ''
            ).geturl()


#
# Copyright (C) 2019 IHS Markit.
# All Rights Reserved
#

from ..arrow_service.arrow_service import ArrowService
from ..arrow_service.exceptions import NoFilesInDataset, UnsupportedDatasetFormat

__all__ = ['ArrowService', 'NoFilesInDataset', 'UnsupportedDatasetFormat']




class AnalyticsAspect:

    # N.B. The contract is that the wrapped_object HAS_A _analytics_handler property, and logger property

    @staticmethod
    def invoke_pre_call_aspects(wrapped_object, metadata):
        pass

    def invoke_post_call_aspects(self, wrapped_object, metadata):
        try:
            additional_properties = metadata.get('properties', {})
            self._create_event(wrapped_object, metadata, status_code=200, additional_properties=additional_properties)
        except Exception as e:
            wrapped_object.logger.exception(
                'Error while invoking post-call aspects.', e
            )

    def invoke_after_exception_aspects(self, wrapped_object, metadata, exception):
        try:
            status_code = self._retrieve_status_code_from_exception(exception)
            additional_properties = metadata.get('properties', {})
            self._create_event(
                wrapped_object, metadata,
                status_code=status_code,
                additional_properties=additional_properties)
        except Exception as e:
            wrapped_object.logger.exception(
                'Error while invoking after-exception aspects.', e
            )

    @staticmethod
    def _create_event(wrapped_object, metadata, status_code, additional_properties):
        if getattr(wrapped_object, '_analytics_handler', None):
            wrapped_object._analytics_handler.create_event(
                metadata['subject'], metadata['organisation_id'],
                metadata['func'].__qualname__.split('.')[0],
                metadata['func'].__name__,
                {**metadata['arguments'], **metadata['kwargs'], **additional_properties},
                result_status_code=status_code
            )

    @staticmethod
    def _retrieve_status_code_from_exception(exception):
        try:
            return exception.response.status_code
        except AttributeError:
            return 500



import contextlib

from platform_services_lib.lib.context.urls import dataset_urls
from platform_services_lib.lib.services.dlc_attributes_dict import AttributesDict


class SampleDataModel:
    def __init__(self, parent):
        self._parent = parent
        self._client = parent._client

    def schema(self):
        """
        Returns the schema data and first rows of sample data.

        :returns: attributes dictionary
        """
        response = self._client.session.get(
            dataset_urls.v2_sample_data_schema.format(id=self._parent.id)
        )

        return AttributesDict(**response.json()['data']['attributes'])

    @contextlib.contextmanager
    def file(self):
        """
        Get the sample data.
        The response type varies depending on whether the dataset is setup to return a static
        preview or a live preview.
        If the preview type is a static preview, then this function provides a file like object
        containing sample data.

        Example usage for static preview of a .parquet format file:

        .. code-block:: python

            >>> import pandas
            >>> import dli

            >>> client = dli.connect()

            >>> dataset = client.datasets.get('some-dataset-short-code')

            >>> with dataset.sample_data.file() as f:
            >>>    dataframe = pandas.read_parquet(f)

        Example usage for static preview of a .csv format file:

        .. code-block:: python

            >>> import pandas
            >>> import dli

            >>> client = dli.connect()

            >>> dataset = client.datasets.get('some-dataset-short-code')

            >>> with dataset.sample_data.file() as f:
            >>>    dataframe = pandas.read_csv(f)

        Else if the preview type is a live preview, then this function provides a HTTP response
        that needs to be wrapped for pandas to be able to read

        Example usage for live preview of a .parquet format file:

        .. code-block:: python

            >>> import pandas
            >>> import dli
            >>> import io

            >>> client = dli.connect()

            >>> dataset = client.datasets.get('some-dataset-short-code')

            >>> with dataset.sample_data.file() as f:
            >>>     dataframe = pd.read_parquet(io.BytesIO(f.read()))

        Example usage for live preview of a .parquet format file:

        .. code-block:: python

            >>> import pandas
            >>> import dli
            >>> import io

            >>> client = dli.connect()

            >>> dataset = client.datasets.get('some-dataset-short-code')

            >>> with dataset.sample_data.file() as f:
            >>>     dataframe = pd.read_csv(io.BytesIO(f.read()))
        """
        response = self._client.session.get(
            dataset_urls.v2_sample_data_file.format(id=self._parent.id),
            stream=True
        )
        # otherwise you get raw secure
        response.raw.decode_content = True
        yield response.raw
        response.close()




from collections import UserDict
from typing import Mapping


class AttributesDict(UserDict):

    def __init__(self, *args, **kwargs):
        # recursively provide the rather silly attribute access
        data = {}

        for arg in args:
            data.update(arg)

        data.update(**kwargs)

        for key, value in data.items():
            if isinstance(value, Mapping):
                self.__dict__[key] = AttributesDict(value)
            else:
                self.__dict__[key] = value

    def __getitem__(self, key):
        return self.__dict__[key]

    def __iter__(self):
        return iter(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def __repr__(self):
        attributes = ' '.join([
            '{}={}'.format(k, v) for k,v in self.__dict__.items()
            if not k.startswith('_')
        ])

        return "{}({})".format(self.__class__.__name__, attributes)



import logging
import operator
import re
from typing import Optional, List
from dateutil.parser import isoparse

from ..services.exceptions import UnableToParsePartition

logger = logging.getLogger(__name__)

class PartitionOperatorValueModel:
    # todo - remove this model, its only used then jettisoned for the as_dict result
    """
    Field that serializes to a title case string and deserializes to a lower case string.
    """
    _re_match = re.compile(
        r'^([\sa-zA-z0-9_-]+)'
        '(<=|<|=|>|>=|!=)((?!<=|<|=|>|>=|!=)'
        r'[\sa-zA-z0-9_-]+)$'
    )

    def __init__(self, val):
        matches = PartitionOperatorValueModel._re_match.match(val)
        self.valid = True
        if not matches or len(matches.groups()) != 3:
            self.valid = False
            raise ValueError(
                f"Requested partition is invalid: {val}. Partition arguments "
                f"must be alphanumeric separated by an operator, and must "
                f"not be wrapped in special characters like single or double "
                f"quotations."
            )
        else:
            self.partition, self.operator, self.value = matches.groups()

    def as_dict(self):
        if self.valid:
            return {
                'partition': str(self.partition),
                'operator': str(self.operator),
                'value': str(self.value)
            }

        return None


def _operator_lookup(op):
    return {
        '<': operator.lt,
        '>': operator.gt,
        '=': operator.eq,
        '<=': operator.le,
        '>=': operator.ge,
        '!=': operator.ne
    }[op]


def _eval_logical(oper, field, val):
    return _operator_lookup(oper)(field, val)


def get_partitions_in_filepath(filepath: str):
    splits = filepath.split("/")
    # n.b. kind of unsafe as need specifically two elements to each sub-list else will break
    return [x.split("=") for x in splits if "=" in x]


def meets_partition_params(
        file_path: str, query_partitions: Optional[List]
):
    if query_partitions is None:
        return True

    found_partitions = dict(get_partitions_in_filepath(file_path))

    filtered = [
        x for x in query_partitions
        if x['partition'] in found_partitions
    ]

    # Example:
    # found_p = dict[(k,v), (k1,v1)] = k:v, k1:v1
    # query_p = [{'partition':'date', 'operator':'<', 'value':'20190102'}]

    for filterable in filtered:
        field = filterable['partition'].strip()
        compare_val = found_partitions[field]
        op = filterable['operator'].strip()
        filter_value = filterable['value'].strip()

        if field.startswith('as_of_'):
            try:
                filter_value_date = isoparse(filter_value)
            except ValueError as e:
                # This means a user has given an invalid date i.e. not in
                # the correct ISO 8601 format for Python datetime.
                raise UnableToParsePartition(
                    f'Was unable to parse the filter value provided: '
                    f'{filter_value} file_path: {file_path}'
                ) from e

            try:
                compare_val = isoparse(compare_val)
            except ValueError:
                # Can not meet partition params as it not a date
                logger.warning(
                    f'{file_path} is not a valid date.',
                    extra={
                        'file_path': file_path,
                        'filterable': filterable
                    },
                    exc_info=True
                )

                return False

            filter_value = filter_value_date

        match = _eval_logical(
            op,
            field=compare_val,
            val=filter_value
        )

        if not match:
            # Short circuit as soon as we fail to match a filter.
            logger.debug(
                f"Excluding file with path '{file_path}' because it "
                f"contains the partitions '{found_partitions}' and the "
                f"user is filtering with '{field}{op}{filter_value}' "
                f"which for this path is {compare_val}'."
            )
            return False

    not_filtered = [
        x for x in query_partitions
        if not x['partition'] in found_partitions
    ]

    if not_filtered:
        logger.warning(
            f"These query partitions '{not_filtered}' were not found as "
            f"keys in the S3 path '{file_path}', so we are going to "
            'let this file through the filter but either the user has '
            'supplied an partition that does not exist or one of the S3 '
            'paths does not follow the partition pattern of the first S3 '
            'path in this instance.'
        )

    return True


def match_partitions(
    file_path: str,
    partitions: List[str],
):
    """
    Return True if the path contains the partition(s) provided.

    :param file_path:
    :param partitions:
    :return:
    """
    # Convert from a list of string to a dictionary of partition, operator &
    # value
    query_partitions = [
        potential.as_dict() for potential in
        [
            PartitionOperatorValueModel(x)
            for x in partitions
        ] if potential.valid
    ]

    return meets_partition_params(file_path, query_partitions)


import queue
import threading
from typing import Callable, Optional
from urllib.parse import urljoin

import requests

from ..context.urls import consumption_urls
from abc import ABC


class AnalyticsSender(ABC, threading.Thread):

    BUFFER_THRESHOLD = 10

    def __init__(self, config, session_func: Optional[Callable[[], requests.Session]]):
        self.__session_func = session_func
        super().__init__()
        self.setDaemon(True)
        self.config = config
        self.queue = queue.Queue()
        self._buffer = []
        self.start()

    def run(self):
        while True:
            # Send buffer every 5 seconds
            try:
                event = self.queue.get(block=True, timeout=5)
                if event is None:
                    # end thread
                    break

                self._buffer.append(event)

                if len(self._buffer) > self.BUFFER_THRESHOLD:
                    self._send_buffer()
            except queue.Empty:
                # Means we've timed out.
                # send what's in the buffer
                self._send_buffer()
                continue

    def _send_buffer(self, timeout: float = 5):
        """
        Send any queued analytics events to our analytics backend in a single request. If there is any problem with
        sending then we are keeping the code simple and dropping the event (many endpoints on the backend also trigger
        analytics).

        :param float timeout: Default 5. How long to wait for the server to send data before giving up. If this is
            too low (e.g. 0.1) then we give up before the connection is established! 0.5 works in a VirtualBox VM, but
            we have seen some users on slow networks (either because they are in a different regions or because of
            firewalls) so we can set this value high as it happens in a background thread. 5 was chosen to match the
            polling time for the analytics queue.
        """
        try:
            if len(self._buffer):
                url = urljoin(
                    self.config.get('consumption_env'),
                    consumption_urls.consumption_analytics
                )
                if self.__session_func:
                    # Use a function to get the session. The session cannot be passed by reference as it will need to
                    # be refreshed automatically when it expires.
                    session = self.__session_func()
                    # Check session exists, as we do not have a session before the `dli.connect` has been called.
                    if session:
                        session.post(url, json={'data': self._buffer}, timeout=timeout)
                self._buffer = []

        except Exception as e:
            # Data scientists do not want to see stack dumps by default,
            # especially when we have a root cause that triggers secondary
            # exceptions.
            if self.config.get('strict') and self.config.get('logger'):
                logger = self.config.get('logger')
                logger.exception(
                    'Error while sending analytics: ', e
                )


class AnalyticsHandler:
    __acceptable_properties = [
        'package_id', 'dataset_id', 'name', 'dataset_name', 'datafile_id',
        'dictionary_id', 'api_key'
    ]

    def __init__(self, analytics_sender: AnalyticsSender):
        self._analytics_sender = analytics_sender

    def create_event(
        self, user_id, organisation_id,
        entity=None, action=None, properties=None, result_status_code=None
    ):
        body = self._prepare_body_to_send(
            user_id, organisation_id, entity, action, properties,
            result_status_code,
        )
        self._send_event_to_analytics(body)

    def _prepare_body_to_send(
        self, user_id, organisation_id, entity, action, properties, result
    ):
        properties_to_send = self._filter_out_properties(properties)
        properties_to_send = self._override_properties_for_dataset_functions(
            action, entity, properties, properties_to_send)

        event = {
            'application_name': self._analytics_sender.config.get('app_name', 'Unknown'),
            'application_version': self._analytics_sender.config.get('version', 'Unknown'),
            'user_id': user_id,
            'entity': entity,
            'action': action,
            'organisation_id': organisation_id,
            'result': result,
            'properties': properties_to_send
        }
        return {'attributes': event}

    def _filter_out_properties(self, properties):
        return {
            k: v
            for k, v in properties.items() if k in self.__acceptable_properties
        }

    # This is ugly, but 1) register_dataset accepts builder,
    # so we need to extract data from it 2) some functions accept 'id'
    # instead of dataset_id. We either ovveride it like this or not make it
    # generic at all.
    def _override_properties_for_dataset_functions(
            self, action, entity, properties, props
    ):
        if entity == 'Dataset':
            if action == 'register_dataset':
                props = self._filter_out_properties(properties['builder']._data)
            elif 'id' in properties:
                props['dataset_id'] = properties['id']
        return props

    def __del__(self):
        # tell the queue it's over
        self._analytics_sender.queue.put(None)

    def _send_event_to_analytics(self, body):
        # if not os.environ.get('DEV_MODE', 'False').lower() == 'true':
        self._analytics_sender.queue.put(body)


#  #
#  Copyright (C) 2020 IHS Markit.
#  All Rights Reserved
#  #
#

import logging
import socket
import os
import elasticache_auto_discovery
from pymemcache.client.hash import HashClient
from datetime import datetime, timedelta
from pymemcache import serde


logger = logging.getLogger(__name__)


def get_or_refresh_hashclient(func):
    def wrapper(self, *args, **kwargs):
        if not self._client or self._should_rediscover():
            self._rediscover_and_recreate_client_on_new_config()
        return func(self, *args, **kwargs)
    return wrapper


class RefreshingHashClient:
    def __init__(self, config):
        self.config = config
        self._client = None
        self._discovered_nodes = []
        self._last_successful_discovery = datetime.min

    @get_or_refresh_hashclient
    def get(self, key):
        try:
            return self._client.get(key, None)
        finally:
            return None

    @get_or_refresh_hashclient
    def set(self, key, value, expire=0):
        try:
            self._client.set(key, value, expire=expire, noreply=False)
        finally:
            return None

    @get_or_refresh_hashclient
    def delete(self, key):
        try:
            self._client.delete(key, noreply=False)
        finally:
            return None

    def _rediscover_and_recreate_client_on_new_config(self):
        nodes = self._discover_nodes()
        if nodes and nodes != self._discovered_nodes:
            self._discovered_nodes = nodes
            self._client = \
                HashClient(
                    servers=self._discovered_nodes,
                    use_pooling=True,
                    connect_timeout=self.config.time_to_timeout,
                    serde=serde.pickle_serde,
                )

    def _should_rediscover(self):
        return datetime.utcnow() - self._last_successful_discovery > timedelta(
            minutes=5.0
        )

    def _discover_nodes(self):
        # if running local you will not be able to reach EC memcache
        # but you must have a local memcache running (e.g. localhost:11211 - see /etc/memcached.conf)
        # start it with (sudo apt-get install memcached; sudo systemctl start/stop memcached)
        if os.environ.get('LOCAL_MEMCACHE', None):
            logger.warning(f"EC Memcache unreachable, connecting to local memcache {os.environ.get('LOCAL_MEMCACHE')}")
            return [os.environ.get('LOCAL_MEMCACHE').split(":")]

        try:
            nodes = elasticache_auto_discovery.discover(
                self.config.elasticache_config_node_url,
                time_to_timeout=self.config.time_to_timeout
            )
        except socket.timeout as s:

            logger.warning(f"Memcache is disabled because an exception occurred: {s}")
            return None
        except Exception as e:
            logger.warning(f"Memcache is disabled because an exception occurred: {e}")
            return None

        nodes = [
            (ip.decode("utf-8"), int(port)) for _server, ip, port in nodes
        ]

        logger.info(
            'elasticache_auto_discovery nodes',
            extra={
                'nodes': str(nodes),
                'num nodes': len(nodes)
            }
        )
        return nodes



#
# Copyright (C) 2019 IHS Markit.
# All Rights Reserved
#

from typing import Optional
from collections import namedtuple
try:
    from injector import Module
except ModuleNotFoundError:
    Module = object  # type: ignore

Dependency = namedtuple(
    'Dependency',
    ['binding_key', 'method', 'scope']
)


class BaseModule(Module):
    """
    Defines a injector module with some predefined lifecyle methods.
    Please make future dependencies inherit from this class as it
    greatly simplifies testing.

    ----

    Supporting flask extensions can be done like this.

    https://github.com/alecthomas/flask_injector#supporting-flask-extensions
    """
    config_class: Optional[type] = None

    def __init__(self, **kwargs):
        super().__init__()
        if self.config_class:
            self.config = self.config_class(**kwargs)



#
# Copyright (C) 2020 IHS Markit.
# All Rights Reserved
#
import os
from distutils.version import StrictVersion
import warnings
import requests
import logging

from dli.client.logging import _setup_logging
from dli.client.dli_client import DliClient
from dli import __version__


logger = logging.getLogger(__name__)
trace_logger = logging.getLogger('trace_logger')
DEPRECATED_AGE = 1


def _set_pypi_url(package_name):
    return f"https://pypi.org/pypi/{package_name}/json"


def _is_release(x):
    return x.replace(".", "").isnumeric()


def _version_check(package_name):
    url = _set_pypi_url(package_name)

    try:
        data = requests.get(url).json()
        versions = list(x for x in data["releases"].keys() if _is_release(x))
        versions.sort(key=StrictVersion)
        versions.reverse()

        offset = versions.index(__version__)
        if offset > DEPRECATED_AGE:
            warnings.warn("You are using an old version of the SDK, please "
                          "upgrade using `pip install dli --upgrade` "
                          "before the SDK no longer functions as expected",
                          PendingDeprecationWarning)

    except requests.exceptions.SSLError as e:
        # Version check has been see to fail by Gilberto with an SSLError
        # when doing the handshake with pypi.
        trace_logger.warning(
            f'Could not connect to pypi to check the installed DLI '
            f'version. This may be due to your internet '
            f'security rules. You can check for the latest version by '
            f'visiting https://pypi.org/project/dli/',
            extra={'SSLError': e}
        )
        pass
    except Exception:
        pass


def _start_session(
    root_url="https://catalogue.datalake.ihsmarkit.com/__api",
    debug=False,
    strict=True,
    use_keyring=True,
    log_level=None,
    show_login_url=False,
):
    if log_level is None:
        log_level = 'stderr:info'
    logger = _setup_logging(log_level)

    _version_check('dli')

    user = os.environ.get("DLI_ACCESS_KEY_ID")
    pasw = os.environ.get("DLI_SECRET_ACCESS_KEY")

    logger.debug("Checking auth flow")

    try:
        if user is not None and pasw is not None:
            logger.debug("Using SAM client credentials authentication flow")
            return get_client()(
                root_url,
                debug=debug,
                strict=strict,
                access_id=user, secret_key=pasw,
                use_keyring=use_keyring,
            )
        elif user is None or pasw is None:
            logger.debug("Using SAM PKCE web authentication flow")
            return get_client()(
                root_url,
                debug=debug,
                strict=strict,
                use_keyring=use_keyring,
                show_login_url=show_login_url
            )
    except SystemExit:
        return


def get_client():
    return DliClient
